from __future__ import annotations

import warnings
from datetime import datetime
from typing import Dict, List, Optional, Sequence, Tuple, Type, Union

from typing_extensions import Self, TypeAlias, assert_never

import datahub.metadata.schema_classes as models
from datahub.cli.cli_utils import first_non_null
from datahub.emitter.mce_builder import DEFAULT_ENV
from datahub.errors import (
    IngestionAttributionWarning,
    ItemNotFoundError,
    SchemaFieldKeyError,
    SdkUsageError,
)
from datahub.ingestion.source.sql.sql_types import resolve_sql_type
from datahub.metadata.urns import DatasetUrn, SchemaFieldUrn, Urn
from datahub.sdk._attribution import is_ingestion_attribution
from datahub.sdk._shared import (
    DatasetUrnOrStr,
    DomainInputType,
    HasContainer,
    HasDomain,
    HasInstitutionalMemory,
    HasOwnership,
    HasPlatformInstance,
    HasStructuredProperties,
    HasSubtype,
    HasTags,
    HasTerms,
    LinksInputType,
    OwnersInputType,
    ParentContainerInputType,
    StructuredPropertyInputType,
    TagInputType,
    TagsInputType,
    TermInputType,
    TermsInputType,
    make_time_stamp,
    parse_time_stamp,
)
from datahub.sdk._utils import add_list_unique, remove_list_unique
from datahub.sdk.entity import Entity, ExtraAspectsType
from datahub.utilities.sentinels import Unset, unset

SchemaFieldInputType: TypeAlias = Union[
    Tuple[str, str],  # (name, type)
    Tuple[str, str, str],  # (name, type, description)
    models.SchemaFieldClass,
]
SchemaFieldsInputType: TypeAlias = Union[
    Sequence[SchemaFieldInputType],
    models.SchemaMetadataClass,
]

UpstreamInputType: TypeAlias = Union[
    # Dataset upstream variants.
    DatasetUrnOrStr,
    models.UpstreamClass,
    # Column upstream variants.
    models.FineGrainedLineageClass,
]
# Mapping of { downstream_column -> [upstream_columns] }
ColumnLineageMapping: TypeAlias = Dict[str, List[str]]
UpstreamLineageInputType: TypeAlias = Union[
    models.UpstreamLineageClass,
    List[UpstreamInputType],
    # Combined variant.
    # Map of { upstream_dataset -> { downstream_column -> [upstream_column] } }
    Dict[DatasetUrnOrStr, ColumnLineageMapping],
]


def _parse_upstream_input(
    upstream_input: UpstreamInputType,
) -> Union[models.UpstreamClass, models.FineGrainedLineageClass]:
    if isinstance(
        upstream_input, (models.UpstreamClass, models.FineGrainedLineageClass)
    ):
        return upstream_input
    elif isinstance(upstream_input, (str, DatasetUrn)):
        return models.UpstreamClass(
            dataset=str(upstream_input),
            type=models.DatasetLineageTypeClass.TRANSFORMED,
        )
    else:
        assert_never(upstream_input)


def parse_cll_mapping(
    *,
    upstream: DatasetUrnOrStr,
    downstream: DatasetUrnOrStr,
    cll_mapping: ColumnLineageMapping,
) -> List[models.FineGrainedLineageClass]:
    cll = []
    for downstream_column, upstream_columns in cll_mapping.items():
        cll.append(
            models.FineGrainedLineageClass(
                upstreamType=models.FineGrainedLineageUpstreamTypeClass.FIELD_SET,
                downstreamType=models.FineGrainedLineageDownstreamTypeClass.FIELD,
                upstreams=[
                    SchemaFieldUrn(upstream, upstream_column).urn()
                    for upstream_column in upstream_columns
                ],
                downstreams=[SchemaFieldUrn(downstream, downstream_column).urn()],
            )
        )
    return cll


def _parse_upstream_lineage_input(
    upstream_input: UpstreamLineageInputType, downstream_urn: DatasetUrn
) -> models.UpstreamLineageClass:
    if isinstance(upstream_input, models.UpstreamLineageClass):
        return upstream_input
    elif isinstance(upstream_input, list):
        upstreams = [_parse_upstream_input(upstream) for upstream in upstream_input]

        # Partition into table and column lineages.
        tll = [
            upstream
            for upstream in upstreams
            if isinstance(upstream, models.UpstreamClass)
        ]
        cll = [
            upstream
            for upstream in upstreams
            if not isinstance(upstream, models.UpstreamClass)
        ]

        # TODO: check that all things in cll are also in tll
        return models.UpstreamLineageClass(upstreams=tll, fineGrainedLineages=cll)
    elif isinstance(upstream_input, dict):
        tll = []
        cll = []
        for dataset_urn, column_lineage in upstream_input.items():
            tll.append(
                models.UpstreamClass(
                    dataset=str(dataset_urn),
                    type=models.DatasetLineageTypeClass.TRANSFORMED,
                )
            )
            cll.extend(
                parse_cll_mapping(
                    upstream=dataset_urn,
                    downstream=downstream_urn,
                    cll_mapping=column_lineage,
                )
            )

        return models.UpstreamLineageClass(upstreams=tll, fineGrainedLineages=cll)
    else:
        assert_never(upstream_input)


class SchemaField:
    __slots__ = ("_parent", "_field_path")

    def __init__(self, parent: Dataset, field_path: str):
        self._parent = parent
        self._field_path = field_path

    def _base_schema_field(self) -> models.SchemaFieldClass:
        # This must exist - if it doesn't, we've got a larger bug.
        schema_dict = self._parent._schema_dict()
        return schema_dict[self._field_path]

    def _get_editable_schema_field(
        self,
    ) -> Optional[models.EditableSchemaFieldInfoClass]:
        # This method does not make any mutations.
        editable_schema = self._parent._get_aspect(models.EditableSchemaMetadataClass)
        if editable_schema is None:
            return None
        for field in editable_schema.editableSchemaFieldInfo:
            if field.fieldPath == self._field_path:
                return field
        return None

    def _ensure_editable_schema_field(self) -> models.EditableSchemaFieldInfoClass:
        if is_ingestion_attribution():
            warnings.warn(
                "This method should not be used in ingestion mode.",
                IngestionAttributionWarning,
                stacklevel=2,
            )
        editable_schema = self._parent._setdefault_aspect(
            models.EditableSchemaMetadataClass(editableSchemaFieldInfo=[])
        )
        for field in editable_schema.editableSchemaFieldInfo:
            if field.fieldPath == self._field_path:
                return field

        # If we don't have an entry for this field yet, create one.
        field = models.EditableSchemaFieldInfoClass(fieldPath=self._field_path)
        editable_schema.editableSchemaFieldInfo.append(field)
        return field

    @property
    def field_path(self) -> str:
        return self._field_path

    @property
    def mapped_type(self) -> models.SchemaFieldDataTypeClass:
        return self._base_schema_field().type

    @property
    def native_type(self) -> str:
        return self._base_schema_field().nativeDataType

    # TODO expose nullability and primary/foreign key details

    @property
    def description(self) -> Optional[str]:
        editable_field = self._get_editable_schema_field()
        return first_non_null(
            [
                editable_field.description if editable_field is not None else None,
                self._base_schema_field().description,
            ]
        )

    def set_description(self, description: str) -> None:
        if is_ingestion_attribution():
            editable_field = self._get_editable_schema_field()
            if editable_field and editable_field.description is not None:
                warnings.warn(
                    "The field description will be hidden by UI-based edits. "
                    "Change the edit mode to OVERWRITE_UI to override this behavior.",
                    category=IngestionAttributionWarning,
                    stacklevel=2,
                )

            self._base_schema_field().description = description
        else:
            self._ensure_editable_schema_field().description = description

    @property
    def tags(self) -> Optional[List[models.TagAssociationClass]]:
        # Tricky: if either has a non-null globalTags, this will not return None.
        tags = None

        if (base_tags := self._base_schema_field().globalTags) is not None:
            tags = tags or []
            tags.extend(base_tags.tags)

        if editable_field := self._get_editable_schema_field():
            if (editable_tags := editable_field.globalTags) is not None:
                tags = tags or []
                tags.extend(editable_tags.tags)

        return tags

    def set_tags(self, tags: TagsInputType) -> None:
        parsed_tags = [self._parent._parse_tag_association_class(tag) for tag in tags]

        if is_ingestion_attribution():
            editable_field = self._get_editable_schema_field()
            if editable_field and editable_field.globalTags:
                warnings.warn(
                    "Overwriting non-ingestion tags from ingestion is an anti-pattern.",
                    category=IngestionAttributionWarning,
                    stacklevel=2,
                )
                editable_field.globalTags = None

            self._base_schema_field().globalTags = models.GlobalTagsClass(
                tags=parsed_tags
            )
        else:
            base_field = self._base_schema_field()
            if base_field.globalTags:
                base_field.globalTags = None

            self._ensure_editable_schema_field().globalTags = models.GlobalTagsClass(
                tags=parsed_tags
            )

    def add_tag(self, tag: TagInputType) -> None:
        parsed_tag = self._parent._parse_tag_association_class(tag)

        if is_ingestion_attribution():
            raise SdkUsageError(
                "Adding field tags in ingestion mode is not yet supported. "
                "Use set_tags instead."
            )
        else:
            editable_field = self._ensure_editable_schema_field()
            if editable_field.globalTags is None:
                editable_field.globalTags = models.GlobalTagsClass(tags=[])

            add_list_unique(
                editable_field.globalTags.tags,
                key=self._parent._tag_key,
                item=parsed_tag,
            )

    def remove_tag(self, tag: TagInputType) -> None:
        parsed_tag = self._parent._parse_tag_association_class(tag)

        if is_ingestion_attribution():
            raise SdkUsageError(
                "Adding field tags in ingestion mode is not yet supported. "
                "Use set_tags instead."
            )
        else:
            base_field = self._base_schema_field()
            if base_field.globalTags is not None:
                remove_list_unique(
                    base_field.globalTags.tags,
                    key=self._parent._tag_key,
                    item=parsed_tag,
                    missing_ok=True,
                )

            editable_field = self._ensure_editable_schema_field()
            if editable_field.globalTags is not None:
                remove_list_unique(
                    editable_field.globalTags.tags,
                    key=self._parent._tag_key,
                    item=parsed_tag,
                )

    @property
    def terms(self) -> Optional[List[models.GlossaryTermAssociationClass]]:
        # TODO: Basically the same implementation as tags - can we share code?
        terms = None

        if (base_terms := self._base_schema_field().glossaryTerms) is not None:
            terms = terms or []
            terms.extend(base_terms.terms)

        if editable_field := self._get_editable_schema_field():
            if (editable_terms := editable_field.glossaryTerms) is not None:
                terms = terms or []
                terms.extend(editable_terms.terms)

        return terms

    def set_terms(self, terms: TermsInputType) -> None:
        parsed_terms = [
            self._parent._parse_glossary_term_association_class(term) for term in terms
        ]

        if is_ingestion_attribution():
            editable_field = self._get_editable_schema_field()
            if editable_field and editable_field.glossaryTerms:
                warnings.warn(
                    "Overwriting non-ingestion terms from ingestion is an anti-pattern.",
                    category=IngestionAttributionWarning,
                    stacklevel=2,
                )
                editable_field.glossaryTerms = None

            self._base_schema_field().glossaryTerms = models.GlossaryTermsClass(
                terms=parsed_terms,
                auditStamp=self._parent._terms_audit_stamp(),
            )
        else:
            base_field = self._base_schema_field()
            if base_field.glossaryTerms:
                base_field.glossaryTerms = None

            self._ensure_editable_schema_field().glossaryTerms = (
                models.GlossaryTermsClass(
                    terms=parsed_terms,
                    auditStamp=self._parent._terms_audit_stamp(),
                )
            )

    def add_term(self, term: TermInputType) -> None:
        parsed_term = self._parent._parse_glossary_term_association_class(term)

        if is_ingestion_attribution():
            raise SdkUsageError(
                "Adding field terms in ingestion mode is not yet supported. "
                "Use set_terms instead."
            )
        else:
            editable_field = self._ensure_editable_schema_field()
            if editable_field.glossaryTerms is None:
                editable_field.glossaryTerms = models.GlossaryTermsClass(
                    terms=[],
                    auditStamp=self._parent._terms_audit_stamp(),
                )

            add_list_unique(
                editable_field.glossaryTerms.terms,
                key=self._parent._terms_key,
                item=parsed_term,
            )

    def remove_term(self, term: TermInputType) -> None:
        parsed_term = self._parent._parse_glossary_term_association_class(term)

        if is_ingestion_attribution():
            raise SdkUsageError(
                "Removing field terms in ingestion mode is not yet supported. "
                "Use set_terms instead."
            )
        else:
            base_field = self._base_schema_field()
            if base_field.glossaryTerms is not None:
                remove_list_unique(
                    base_field.glossaryTerms.terms,
                    key=self._parent._terms_key,
                    item=parsed_term,
                    missing_ok=True,
                )

            editable_field = self._ensure_editable_schema_field()
            if editable_field.glossaryTerms is not None:
                remove_list_unique(
                    editable_field.glossaryTerms.terms,
                    key=self._parent._terms_key,
                    item=parsed_term,
                    missing_ok=True,
                )


class Dataset(
    HasPlatformInstance,
    HasSubtype,
    HasContainer,
    HasOwnership,
    HasInstitutionalMemory,
    HasTags,
    HasTerms,
    HasDomain,
    HasStructuredProperties,
    Entity,
):
    """Represents a dataset in DataHub.

    A dataset represents a collection of data, such as a table, view, or file.
    This class provides methods for managing dataset metadata including schema,
    lineage, and various aspects like ownership, tags, and terms.
    """

    __slots__ = ()

    @classmethod
    def get_urn_type(cls) -> Type[DatasetUrn]:
        """Get the URN type for datasets.

        Returns:
            The DatasetUrn class.
        """
        return DatasetUrn

    def __init__(
        self,
        *,
        # Identity.
        platform: str,
        name: str,
        platform_instance: Optional[str] = None,
        env: str = DEFAULT_ENV,
        # Dataset properties.
        description: Optional[str] = None,
        display_name: Optional[str] = None,
        qualified_name: Optional[str] = None,
        external_url: Optional[str] = None,
        custom_properties: Optional[Dict[str, str]] = None,
        created: Optional[datetime] = None,
        last_modified: Optional[datetime] = None,
        # Standard aspects.
        parent_container: ParentContainerInputType | Unset = unset,
        subtype: Optional[str] = None,
        owners: Optional[OwnersInputType] = None,
        links: Optional[LinksInputType] = None,
        tags: Optional[TagsInputType] = None,
        terms: Optional[TermsInputType] = None,
        domain: Optional[DomainInputType] = None,
        # Dataset-specific aspects.
        schema: Optional[SchemaFieldsInputType] = None,
        upstreams: Optional[models.UpstreamLineageClass] = None,
        structured_properties: Optional[StructuredPropertyInputType] = None,
        extra_aspects: ExtraAspectsType = None,
    ):
        """Initialize a new Dataset instance.

        Args:
            platform: The platform this dataset belongs to (e.g. "mysql", "snowflake").
            name: The name of the dataset.
            platform_instance: Optional platform instance identifier.
            env: The environment this dataset belongs to (default: DEFAULT_ENV).
            description: Optional description of the dataset.
            display_name: Optional display name for the dataset.
            qualified_name: Optional qualified name for the dataset.
            external_url: Optional URL to external documentation or source.
            custom_properties: Optional dictionary of custom properties.
            created: Optional creation timestamp.
            last_modified: Optional last modification timestamp.
            parent_container: Optional parent container for this dataset.
            subtype: Optional subtype of the dataset.
            owners: Optional list of owners.
            links: Optional list of links.
            tags: Optional list of tags.
            terms: Optional list of glossary terms.
            domain: Optional domain this dataset belongs to.
            extra_aspects: Optional list of additional aspects.
            schema: Optional schema definition for the dataset.
            upstreams: Optional upstream lineage information.
        """
        urn = DatasetUrn.create_from_ids(
            platform_id=platform,
            table_name=name,
            platform_instance=platform_instance,
            env=env,
        )
        super().__init__(urn)
        self._set_extra_aspects(extra_aspects)

        self._set_platform_instance(urn.platform, platform_instance)

        if schema is not None:
            self._set_schema(schema)
        if upstreams is not None:
            self.set_upstreams(upstreams)

        if description is not None:
            self.set_description(description)
        if display_name is not None:
            self.set_display_name(display_name)
        if qualified_name is not None:
            self.set_qualified_name(qualified_name)
        if external_url is not None:
            self.set_external_url(external_url)
        if custom_properties is not None:
            self.set_custom_properties(custom_properties)
        if created is not None:
            self.set_created(created)
        if last_modified is not None:
            self.set_last_modified(last_modified)

        if parent_container is not unset:
            self._set_container(parent_container)
        if subtype is not None:
            self.set_subtype(subtype)
        if owners is not None:
            self.set_owners(owners)
        if links is not None:
            self.set_links(links)
        if tags is not None:
            self.set_tags(tags)
        if terms is not None:
            self.set_terms(terms)
        if domain is not None:
            self.set_domain(domain)
        if structured_properties is not None:
            for key, value in structured_properties.items():
                self.set_structured_property(property_urn=key, values=value)

    @classmethod
    def _new_from_graph(cls, urn: Urn, current_aspects: models.AspectBag) -> Self:
        assert isinstance(urn, DatasetUrn)
        entity = cls(
            platform=urn.platform,
            name=urn.name,
            env=urn.env,
        )
        return entity._init_from_graph(current_aspects)

    @property
    def urn(self) -> DatasetUrn:
        return self._urn  # type: ignore

    def _ensure_dataset_props(self) -> models.DatasetPropertiesClass:
        return self._setdefault_aspect(models.DatasetPropertiesClass())

    def _get_editable_props(self) -> Optional[models.EditableDatasetPropertiesClass]:
        return self._get_aspect(models.EditableDatasetPropertiesClass)

    def _ensure_editable_props(self) -> models.EditableDatasetPropertiesClass:
        # Note that most of the fields in this aspect are not used.
        # The only one that's relevant for us is the description.
        return self._setdefault_aspect(models.EditableDatasetPropertiesClass())

    @property
    def description(self) -> Optional[str]:
        """Get the description of the dataset.

        Returns:
            The description if set, None otherwise.
        """
        editable_props = self._get_editable_props()
        return first_non_null(
            [
                editable_props.description if editable_props is not None else None,
                self._ensure_dataset_props().description,
            ]
        )

    def set_description(self, description: str) -> None:
        """Set the description of the dataset.

        Args:
            description: The description to set.

        Note:
            If called during ingestion, this will warn if overwriting
            a non-ingestion description.
        """
        if is_ingestion_attribution():
            editable_props = self._get_editable_props()
            if editable_props is not None and editable_props.description is not None:
                warnings.warn(
                    "Overwriting non-ingestion description from ingestion is an anti-pattern.",
                    category=IngestionAttributionWarning,
                    stacklevel=2,
                )
                # Force the ingestion description to show up.
                editable_props.description = None

            self._ensure_dataset_props().description = description
        else:
            self._ensure_editable_props().description = description

    @property
    def display_name(self) -> Optional[str]:
        """Get the display name of the dataset.

        Returns:
            The display name if set, None otherwise.
        """
        return self._ensure_dataset_props().name

    def set_display_name(self, display_name: str) -> None:
        """Set the display name of the dataset.

        Args:
            display_name: The display name to set.
        """
        self._ensure_dataset_props().name = display_name

    @property
    def qualified_name(self) -> Optional[str]:
        """Get the qualified name of the dataset.

        Returns:
            The qualified name if set, None otherwise.
        """
        return self._ensure_dataset_props().qualifiedName

    def set_qualified_name(self, qualified_name: str) -> None:
        """Set the qualified name of the dataset.

        Args:
            qualified_name: The qualified name to set.
        """
        self._ensure_dataset_props().qualifiedName = qualified_name

    @property
    def external_url(self) -> Optional[str]:
        """Get the external URL of the dataset.

        Returns:
            The external URL if set, None otherwise.
        """
        return self._ensure_dataset_props().externalUrl

    def set_external_url(self, external_url: str) -> None:
        """Set the external URL of the dataset.

        Args:
            external_url: The external URL to set.
        """
        self._ensure_dataset_props().externalUrl = external_url

    @property
    def custom_properties(self) -> Dict[str, str]:
        """Get the custom properties of the dataset.

        Returns:
            Dictionary of custom properties.
        """
        return self._ensure_dataset_props().customProperties

    def set_custom_properties(self, custom_properties: Dict[str, str]) -> None:
        """Set the custom properties of the dataset.

        Args:
            custom_properties: Dictionary of custom properties to set.
        """
        self._ensure_dataset_props().customProperties = custom_properties

    @property
    def created(self) -> Optional[datetime]:
        """Get the creation timestamp of the dataset.

        Returns:
            The creation timestamp if set, None otherwise.
        """
        return parse_time_stamp(self._ensure_dataset_props().created)

    def set_created(self, created: datetime) -> None:
        """Set the creation timestamp of the dataset.

        Args:
            created: The creation timestamp to set.
        """
        self._ensure_dataset_props().created = make_time_stamp(created)

    @property
    def last_modified(self) -> Optional[datetime]:
        """Get the last modification timestamp of the dataset.

        Returns:
            The last modification timestamp if set, None otherwise.
        """
        return parse_time_stamp(self._ensure_dataset_props().lastModified)

    def set_last_modified(self, last_modified: datetime) -> None:
        self._ensure_dataset_props().lastModified = make_time_stamp(last_modified)

    def _schema_dict(self) -> Dict[str, models.SchemaFieldClass]:
        schema_metadata = self._get_aspect(models.SchemaMetadataClass)
        if schema_metadata is None:
            raise ItemNotFoundError(f"Schema is not set for dataset {self.urn}")
        return {field.fieldPath: field for field in schema_metadata.fields}

    @property
    def schema(self) -> List[SchemaField]:
        # TODO: Add some caching here to avoid iterating over the schema every time.
        """Get the schema fields of the dataset.

        Returns:
            List of SchemaField objects representing the dataset's schema.
        """
        schema_dict = self._schema_dict()
        return [SchemaField(self, field_path) for field_path in schema_dict]

    def _parse_schema_field_input(
        self, schema_field_input: SchemaFieldInputType
    ) -> models.SchemaFieldClass:
        if isinstance(schema_field_input, models.SchemaFieldClass):
            return schema_field_input
        elif isinstance(schema_field_input, tuple):
            # Support (name, type) and (name, type, description) forms
            if len(schema_field_input) == 2:
                name, field_type = schema_field_input
                description = None
            elif len(schema_field_input) == 3:
                name, field_type, description = schema_field_input
            else:
                assert_never(schema_field_input)
            return models.SchemaFieldClass(
                fieldPath=name,
                type=models.SchemaFieldDataTypeClass(
                    resolve_sql_type(
                        field_type,
                        platform=self.urn.get_data_platform_urn().platform_name,
                    )
                    or models.NullTypeClass()
                ),
                nativeDataType=field_type,
                description=description,
            )
        else:
            assert_never(schema_field_input)

    def _set_schema(self, schema: SchemaFieldsInputType) -> None:
        # This method is not public. Ingestion/restatement users should be setting
        # the schema via the constructor. SDK users that got a dataset from the graph
        # probably shouldn't be adding/removing fields ad-hoc. The field-level mutators
        # can be used instead.
        if isinstance(schema, models.SchemaMetadataClass):
            self._set_aspect(schema)
        else:
            parsed_schema = [self._parse_schema_field_input(field) for field in schema]
            self._set_aspect(
                models.SchemaMetadataClass(
                    fields=parsed_schema,
                    # The rest of these fields are not used, and so we can set them to dummy/default values.
                    schemaName="",
                    platform=self.urn.platform,
                    version=0,
                    hash="",
                    platformSchema=models.SchemalessClass(),
                )
            )

    def __getitem__(self, field_path: str) -> SchemaField:
        # TODO: Automatically deal with field path v2?
        """Get a schema field by its path.

        Args:
            field_path: The path of the field to retrieve.

        Returns:
            A SchemaField instance.

        Raises:
            SchemaFieldKeyError: If the field is not found.
        """
        schema_dict = self._schema_dict()
        if field_path not in schema_dict:
            raise SchemaFieldKeyError(f"Field {field_path} not found in schema")
        return SchemaField(self, field_path)

    @property
    def upstreams(self) -> Optional[models.UpstreamLineageClass]:
        return self._get_aspect(models.UpstreamLineageClass)

    def set_upstreams(self, upstreams: UpstreamLineageInputType) -> None:
        self._set_aspect(_parse_upstream_lineage_input(upstreams, self.urn))



import datetime
import functools
import itertools
import logging
import os
import pathlib
import platform
import signal
import subprocess
import sys
import tempfile
import time
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

import click
import click_spinner
import requests
from expandvars import expandvars
from requests_file import FileAdapter

from datahub.cli.config_utils import DATAHUB_ROOT_FOLDER
from datahub.cli.docker_check import (
    DATAHUB_COMPOSE_LEGACY_VOLUME_FILTERS,
    DATAHUB_COMPOSE_PROJECT_FILTER,
    DOCKER_COMPOSE_PROJECT_NAME,
    DockerComposeVersionError,
    QuickstartStatus,
    check_docker_quickstart,
    get_docker_client,
    run_quickstart_preflight_checks,
)
from datahub.cli.quickstart_versioning import QuickstartVersionMappingConfig
from datahub.ingestion.run.pipeline import Pipeline
from datahub.telemetry import telemetry
from datahub.upgrade import upgrade
from datahub.utilities.perf_timer import PerfTimer

logger = logging.getLogger(__name__)
_ClickPositiveInt = click.IntRange(min=1)

NEO4J_AND_ELASTIC_QUICKSTART_COMPOSE_FILE = (
    "docker/quickstart/docker-compose.quickstart.yml"
)
ELASTIC_QUICKSTART_COMPOSE_FILE = (
    "docker/quickstart/docker-compose-without-neo4j.quickstart.yml"
)
NEO4J_AND_ELASTIC_M1_QUICKSTART_COMPOSE_FILE = (
    "docker/quickstart/docker-compose-m1.quickstart.yml"
)
ELASTIC_M1_QUICKSTART_COMPOSE_FILE = (
    "docker/quickstart/docker-compose-without-neo4j-m1.quickstart.yml"
)
CONSUMERS_QUICKSTART_COMPOSE_FILE = (
    "docker/quickstart/docker-compose.consumers.quickstart.yml"
)
ELASTIC_CONSUMERS_QUICKSTART_COMPOSE_FILE = (
    "docker/quickstart/docker-compose.consumers-without-neo4j.quickstart.yml"
)
KAFKA_SETUP_QUICKSTART_COMPOSE_FILE = (
    "docker/quickstart/docker-compose.kafka-setup.quickstart.yml"
)


_QUICKSTART_MAX_WAIT_TIME = datetime.timedelta(minutes=10)
_QUICKSTART_UP_TIMEOUT = datetime.timedelta(seconds=100)
_QUICKSTART_STATUS_CHECK_INTERVAL = datetime.timedelta(seconds=2)


class Architectures(Enum):
    x86 = "x86"
    arm64 = "arm64"
    m1 = "m1"
    m2 = "m2"


@functools.lru_cache
def _docker_subprocess_env() -> Dict[str, str]:
    # platform.machine() is equivalent to `uname -m`, as per https://stackoverflow.com/a/45124927/5004662
    DOCKER_COMPOSE_PLATFORM: str = "linux/" + platform.machine()

    env = {
        **os.environ,
        "DOCKER_BUILDKIT": "1",
    }
    if DOCKER_COMPOSE_PLATFORM:
        env["DOCKER_DEFAULT_PLATFORM"] = DOCKER_COMPOSE_PLATFORM
    return env


@click.group()
def docker() -> None:
    """Helper commands for setting up and interacting with a local
    DataHub instance using Docker."""
    pass


@docker.command()
@upgrade.check_upgrade
@telemetry.with_telemetry()
def check() -> None:
    """Check that the Docker containers are healthy"""
    status = check_docker_quickstart()
    if status.is_ok():
        click.secho("✔ No issues detected", fg="green")
    else:
        raise status.to_exception("The following issues were detected:")


def is_m1() -> bool:
    """Check whether we are running on an M1 machine"""
    try:
        return (
            platform.uname().machine == "arm64" and platform.uname().system == "Darwin"
        )
    except Exception:
        # Catch-all
        return False


def is_arch_m1(arch: Architectures) -> bool:
    return arch in [Architectures.arm64, Architectures.m1, Architectures.m2]


def should_use_neo4j_for_graph_service(graph_service_override: Optional[str]) -> bool:
    if graph_service_override is not None:
        if graph_service_override == "elasticsearch":
            click.echo("Starting with elasticsearch due to graph-service-impl param\n")
            return False
        if graph_service_override == "neo4j":
            click.echo("Starting with neo4j due to graph-service-impl param\n")
            return True
        else:
            click.secho(
                graph_service_override
                + " is not a valid graph service option. Choose either `neo4j` or "
                "`elasticsearch`\n",
                fg="red",
            )
            raise ValueError(f"invalid graph service option: {graph_service_override}")
    with get_docker_client() as client:
        if len(client.volumes.list(filters={"name": "datahub_neo4jdata"})) > 0:
            click.echo(
                "Datahub Neo4j volume found, starting with neo4j as graph service.\n"
                "If you want to run using elastic, run `datahub docker nuke` and re-ingest your data.\n"
            )
            return True

        logger.debug(
            "No Datahub Neo4j volume found, starting with elasticsearch as graph service.\n"
            "To use neo4j as a graph backend, run \n"
            "`datahub docker quickstart --graph-service-impl neo4j`"
            "\nfrom the root of the datahub repo\n"
        )
        return False


def _set_environment_variables(
    version: Optional[str],
    mysql_version: Optional[str],
    mysql_port: Optional[int],
    zk_port: Optional[int],
    kafka_broker_port: Optional[int],
    schema_registry_port: Optional[int],
    elastic_port: Optional[int],
    kafka_setup: Optional[bool],
) -> None:
    if version is not None:
        if not version.startswith("v") and "." in version:
            logger.warning(
                f"Version passed in '{version}' doesn't start with v, substituting with 'v{version}'"
            )
            version = f"v{version}"
        os.environ["DATAHUB_VERSION"] = version
    if mysql_version is not None:
        os.environ["DATAHUB_MYSQL_VERSION"] = mysql_version
    if mysql_port is not None:
        os.environ["DATAHUB_MAPPED_MYSQL_PORT"] = str(mysql_port)

    if zk_port is not None:
        os.environ["DATAHUB_MAPPED_ZK_PORT"] = str(zk_port)

    if kafka_broker_port is not None:
        os.environ["DATAHUB_MAPPED_KAFKA_BROKER_PORT"] = str(kafka_broker_port)

    if schema_registry_port is not None:
        os.environ["DATAHUB_MAPPED_SCHEMA_REGISTRY_PORT"] = str(schema_registry_port)

    if elastic_port is not None:
        os.environ["DATAHUB_MAPPED_ELASTIC_PORT"] = str(elastic_port)
    if kafka_setup:
        os.environ["DATAHUB_PRECREATE_TOPICS"] = "true"


def _get_default_quickstart_compose_file() -> Optional[str]:
    quickstart_folder = Path(DATAHUB_ROOT_FOLDER) / "quickstart"
    try:
        os.makedirs(quickstart_folder, exist_ok=True)
        return f"{quickstart_folder}/docker-compose.yml"
    except Exception as e:
        logger.debug(f"Failed to identify a default quickstart compose file due to {e}")
    return None


def _docker_compose_v2() -> List[str]:
    try:
        # Check for the docker compose v2 plugin.
        compose_version = subprocess.check_output(
            ["docker", "compose", "version", "--short"], stderr=subprocess.STDOUT
        ).decode()
        assert compose_version.startswith("2.") or compose_version.startswith("v2.")
        return ["docker", "compose"]
    except (OSError, subprocess.CalledProcessError, AssertionError):
        # We'll check for docker-compose as well.
        try:
            compose_version = subprocess.check_output(
                ["docker-compose", "version", "--short"], stderr=subprocess.STDOUT
            ).decode()
            if compose_version.startswith("2.") or compose_version.startswith("v2."):
                # This will happen if docker compose v2 is installed in standalone mode
                # instead of as a plugin.
                return ["docker-compose"]

            raise DockerComposeVersionError(
                f"You have docker-compose v1 ({compose_version}) installed, but we require Docker Compose v2. "
                "Please upgrade to Docker Compose v2. "
                "See https://docs.docker.com/compose/compose-v2/ for more information."
            )
        except (OSError, subprocess.CalledProcessError):
            # docker-compose v1 is not installed either.
            raise DockerComposeVersionError(
                "You don't have Docker Compose installed. Please install Docker Compose. See https://docs.docker.com/compose/install/.",
            ) from None


def _attempt_stop(quickstart_compose_file: List[pathlib.Path]) -> None:
    default_quickstart_compose_file = _get_default_quickstart_compose_file()
    compose_files_for_stopping = (
        quickstart_compose_file
        if quickstart_compose_file
        else (
            [pathlib.Path(default_quickstart_compose_file)]
            if default_quickstart_compose_file
            else None
        )
    )
    if compose_files_for_stopping:
        # docker-compose stop
        compose = _docker_compose_v2()
        base_command: List[str] = [
            *compose,
            *itertools.chain.from_iterable(
                ("-f", f"{path}") for path in compose_files_for_stopping
            ),
            "-p",
            DOCKER_COMPOSE_PROJECT_NAME,
        ]
        try:
            logger.debug(f"Executing {base_command} stop")
            subprocess.run(
                [*base_command, "stop"],
                check=True,
                env=_docker_subprocess_env(),
            )
            click.secho("Stopped datahub successfully.", fg="green")
        except subprocess.CalledProcessError:
            click.secho(
                "Error while stopping.",
                fg="red",
            )
        return


def _backup(backup_file: str) -> int:
    resolved_backup_file = os.path.expanduser(backup_file)
    dirname = os.path.dirname(resolved_backup_file)
    logger.info(f"Creating directory {dirname} if it doesn't exist")
    os.makedirs(dirname, exist_ok=True)
    logger.info("Executing backup command")
    result = subprocess.run(
        [
            "bash",
            "-c",
            f"docker exec {DOCKER_COMPOSE_PROJECT_NAME}-mysql-1 mysqldump -u root -pdatahub datahub > {resolved_backup_file}",
        ]
    )
    logger.info(
        f"Backup written to {resolved_backup_file} with status {result.returncode}"
    )
    return result.returncode


def _restore(
    restore_primary: bool,
    restore_indices: Optional[bool],
    primary_restore_file: Optional[str],
) -> int:
    assert restore_primary or restore_indices, (
        "Either restore_primary or restore_indices must be set"
    )
    msg = "datahub> "
    if restore_primary:
        msg += f"Will restore primary database from {primary_restore_file}. "
    if restore_indices is not False:
        msg += (
            f"Will {'also ' if restore_primary else ''}re-build indexes from scratch. "
        )
    else:
        msg += "Will not re-build indexes. "
    msg += "Press y to continue."
    click.confirm(msg, abort=True)
    if restore_primary:
        assert primary_restore_file
        resolved_restore_file = os.path.expanduser(primary_restore_file)
        logger.info(f"Restoring primary db from backup at {resolved_restore_file}")
        assert os.path.exists(resolved_restore_file), (
            f"File {resolved_restore_file} does not exist"
        )
        with open(resolved_restore_file) as fp:
            result = subprocess.run(
                [
                    "bash",
                    "-c",
                    f"docker exec -i {DOCKER_COMPOSE_PROJECT_NAME}-mysql-1 bash -c 'mysql -uroot -pdatahub datahub '",
                ],
                stdin=fp,
                capture_output=True,
            )
        if result.returncode != 0:
            logger.error("Failed to run MySQL restore")
            return result.returncode
        else:
            logger.info("Successfully restored primary backup.")

    # We run restore indices by default on primary restores, and also if the --restore-indices flag is explicitly set
    if restore_indices is not False:
        with tempfile.NamedTemporaryFile() as env_fp:
            env_fp.write(
                expandvars(
                    """
            # Required Environment Variables
EBEAN_DATASOURCE_USERNAME=datahub
EBEAN_DATASOURCE_PASSWORD=datahub
EBEAN_DATASOURCE_HOST=mysql:${DATAHUB_MAPPED_MYSQL_PORT:-3306}
EBEAN_DATASOURCE_URL=jdbc:mysql://mysql:${DATAHUB_MAPPED_MYSQL_PORT:-3306}/datahub?verifyServerCertificate=false&useSSL=true&useUnicode=yes&characterEncoding=UTF-8
EBEAN_DATASOURCE_DRIVER=com.mysql.jdbc.Driver
ENTITY_REGISTRY_CONFIG_PATH=/datahub/datahub-gms/resources/entity-registry.yml

KAFKA_BOOTSTRAP_SERVER=broker:29092
KAFKA_SCHEMAREGISTRY_URL=http://schema-registry:${DATAHUB_MAPPED_SCHEMA_REGISTRY_PORT:-8081}

ELASTICSEARCH_HOST=elasticsearch
ELASTICSEARCH_PORT=${DATAHUB_MAPPED_ELASTIC_PORT:-9200}

#NEO4J_HOST=http://<your-neo-host>:7474
#NEO4J_URI=bolt://<your-neo-host>
#NEO4J_USERNAME=neo4j
#NEO4J_PASSWORD=datahub

DATAHUB_GMS_HOST=datahub-gms
DATAHUB_GMS_PORT=${DATAHUB_MAPPED_GMS_PORT:-8080}

DATAHUB_MAE_CONSUMER_HOST=datahub-gms
DATAHUB_MAE_CONSUMER_PORT=9091

# Optional Arguments

# Uncomment and set these to support SSL connection to Elasticsearch
# ELASTICSEARCH_USE_SSL=
# ELASTICSEARCH_SSL_PROTOCOL=
# ELASTICSEARCH_SSL_SECURE_RANDOM_IMPL=
# ELASTICSEARCH_SSL_TRUSTSTORE_FILE=
# ELASTICSEARCH_SSL_TRUSTSTORE_TYPE=
# ELASTICSEARCH_SSL_TRUSTSTORE_PASSWORD=
# ELASTICSEARCH_SSL_KEYSTORE_FILE=
# ELASTICSEARCH_SSL_KEYSTORE_TYPE=
# ELASTICSEARCH_SSL_KEYSTORE_PASSWORD=
            """
                ).encode("utf-8")
            )
            env_fp.flush()
            if logger.isEnabledFor(logging.DEBUG):
                with open(env_fp.name) as env_fp_reader:
                    logger.debug(f"Env file contents: {env_fp_reader.read()}")

            # continue to issue the restore indices command
            command = (
                "docker pull acryldata/datahub-upgrade:${DATAHUB_VERSION:-head}"
                + f" && docker run --network datahub_network --env-file {env_fp.name} "
                + "acryldata/datahub-upgrade:${DATAHUB_VERSION:-head} -u RestoreIndices -a clean"
            )
            logger.info(f"Running index restore command: {command}")
            result = subprocess.run(
                args=[
                    "bash",
                    "-c",
                    "docker pull acryldata/datahub-upgrade:"
                    + "${DATAHUB_VERSION:-head}"
                    + f" && docker run --network {DOCKER_COMPOSE_PROJECT_NAME}_network --env-file {env_fp.name} "
                    + "acryldata/datahub-upgrade:${DATAHUB_VERSION:-head}"
                    + " -u RestoreIndices -a clean",
                ],
                capture_output=True,
            )
            logger.info(
                f"Index restore command finished with status {result.returncode}"
            )
        if result.returncode != 0:
            logger.info(result.stderr)
        logger.debug(result.stdout)
    return result.returncode


def detect_quickstart_arch(arch: Optional[str]) -> Architectures:
    running_on_m1 = is_m1()
    if running_on_m1:
        click.secho("Detected M1 machine", fg="yellow")

    quickstart_arch = Architectures.x86 if not running_on_m1 else Architectures.arm64
    if arch:
        matched_arch = [a for a in Architectures if arch.lower() == a.value]
        if not matched_arch:
            click.secho(
                f"Failed to match arch {arch} with list of architectures supported {[a.value for a in Architectures]}"
            )
        quickstart_arch = matched_arch[0]
        click.secho(f"Using architecture {quickstart_arch}", fg="yellow")

    return quickstart_arch


@docker.command()
@click.option(
    "--version",
    type=str,
    default="default",
    help="Datahub version to be deployed. If not set, deploy using the defaults from the quickstart compose. Use 'stable' to start the latest stable version.",
)
@click.option(
    "--build-locally",
    type=bool,
    is_flag=True,
    default=False,
    help="Attempt to build the containers locally before starting",
)
@click.option(
    "--pull-images/--no-pull-images",
    type=bool,
    is_flag=True,
    default=True,
    help="Attempt to pull the containers from Docker Hub before starting",
)
@click.option(
    "-f",
    "--quickstart-compose-file",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    default=[],
    multiple=True,
    help="Use a local docker-compose file instead of pulling from GitHub",
)
@click.option(
    "--dump-logs-on-failure",
    type=bool,
    is_flag=True,
    default=False,
    help="If true, the docker-compose logs will be printed to console if something fails",
)
@click.option(
    "--graph-service-impl",
    type=str,
    is_flag=False,
    default=None,
    help="If set, forces docker-compose to use that graph service implementation",
)
@click.option(
    "--mysql-port",
    type=_ClickPositiveInt,
    is_flag=False,
    default=None,
    help="If there is an existing mysql instance running on port 3306, set this to a free port to avoid port conflicts on startup",
)
@click.option(
    "--zk-port",
    type=_ClickPositiveInt,
    is_flag=False,
    default=None,
    help="If there is an existing zookeeper instance running on port 2181, set this to a free port to avoid port conflicts on startup",
)
@click.option(
    "--kafka-broker-port",
    type=_ClickPositiveInt,
    is_flag=False,
    default=None,
    help="If there is an existing Kafka broker running on port 9092, set this to a free port to avoid port conflicts on startup",
)
@click.option(
    "--schema-registry-port",
    type=_ClickPositiveInt,
    is_flag=False,
    default=None,
    help="If there is an existing process running on port 8081, set this to a free port to avoid port conflicts with Kafka schema registry on startup",
)
@click.option(
    "--elastic-port",
    type=_ClickPositiveInt,
    is_flag=False,
    default=None,
    help="If there is an existing Elasticsearch instance running on port 9092, set this to a free port to avoid port conflicts on startup",
)
@click.option(
    "--stop",
    type=bool,
    is_flag=True,
    default=False,
    help="Use this flag to stop the running containers",
)
@click.option(
    "--backup",
    required=False,
    is_flag=True,
    default=False,
    help="Run this to backup a running quickstart instance",
)
@click.option(
    "--backup-file",
    required=False,
    type=click.Path(exists=False, dir_okay=False, readable=True, writable=True),
    default=os.path.expanduser("~/.datahub/quickstart/backup.sql"),
    show_default=True,
    help="Run this to backup data from a running quickstart instance",
)
@click.option(
    "--restore",
    required=False,
    is_flag=True,
    default=False,
    help="Run this to restore a running quickstart instance from a previous backup (see --backup)",
)
@click.option(
    "--restore-file",
    required=False,
    type=str,
    default=os.path.expanduser("~/.datahub/quickstart/backup.sql"),
    help="Set this to provide a custom restore file",
)
@click.option(
    "--restore-indices",
    required=False,
    is_flag=True,
    default=False,
    help="Enable the restoration of indices of a running quickstart instance. Note: Using --restore will automatically restore-indices unless you use the --no-restore-indices flag.",
)
@click.option(
    "--no-restore-indices",
    required=False,
    is_flag=True,
    default=False,
    help="Disables the restoration of indices of a running quickstart instance when used in conjunction with --restore.",
)
@click.option(
    "--standalone_consumers",
    required=False,
    is_flag=True,
    default=False,
    help="Launches MAE & MCE consumers as stand alone docker containers",
)
@click.option(
    "--kafka-setup",
    required=False,
    is_flag=True,
    default=False,
    help="Launches Kafka setup job as part of the compose deployment",
)
@click.option(
    "--arch",
    required=False,
    help="Specify the architecture for the quickstart images to use. Options are x86, arm64, m1 etc.",
)
@upgrade.check_upgrade
@telemetry.with_telemetry(
    capture_kwargs=[
        "version",
        "build_locally",
        "pull_images",
        "stop",
        "backup",
        "restore",
        "restore_indices",
        "standalone_consumers",
        "kafka_setup",
        "arch",
    ]
)
def quickstart(
    version: Optional[str],
    build_locally: bool,
    pull_images: bool,
    quickstart_compose_file: List[pathlib.Path],
    dump_logs_on_failure: bool,
    graph_service_impl: Optional[str],
    mysql_port: Optional[int],
    zk_port: Optional[int],
    kafka_broker_port: Optional[int],
    schema_registry_port: Optional[int],
    elastic_port: Optional[int],
    stop: bool,
    backup: bool,
    backup_file: str,
    restore: bool,
    restore_file: str,
    restore_indices: bool,
    no_restore_indices: bool,
    standalone_consumers: bool,
    kafka_setup: bool,
    arch: Optional[str],
) -> None:
    """Start an instance of DataHub locally using docker-compose.

    This command will automatically download the latest docker-compose configuration
    from GitHub, pull the latest images, and bring up the DataHub system.
    There are options to override the docker-compose config file, build the containers
    locally, and dump logs to the console or to a file if something goes wrong.
    """
    if backup:
        _backup(backup_file)
        return
    if restore or restore_indices or no_restore_indices:
        if not valid_restore_options(restore, restore_indices, no_restore_indices):
            return
        # execute restore
        restore_indices_flag: Optional[bool] = None
        if restore_indices:
            restore_indices_flag = True
        if no_restore_indices:
            restore_indices_flag = False
        _restore(
            restore_primary=restore,
            primary_restore_file=restore_file,
            restore_indices=restore_indices_flag,
        )
        return

    quickstart_arch = detect_quickstart_arch(arch)
    quickstart_versioning = QuickstartVersionMappingConfig.fetch_quickstart_config()
    quickstart_execution_plan = quickstart_versioning.get_quickstart_execution_plan(
        version
    )
    logger.info(f"Using quickstart plan: {quickstart_execution_plan}")

    # Run pre-flight checks.
    with get_docker_client() as client:
        run_quickstart_preflight_checks(client)

    quickstart_compose_file = list(
        quickstart_compose_file
    )  # convert to list from tuple

    auth_resources_folder = Path(DATAHUB_ROOT_FOLDER) / "plugins/auth/resources"
    os.makedirs(auth_resources_folder, exist_ok=True)

    quickstart_compose_file_name = _get_default_quickstart_compose_file()
    if stop:
        _attempt_stop(quickstart_compose_file)
        return
    elif not quickstart_compose_file:
        logger.info(f"compose file name {quickstart_compose_file_name}")
        download_compose_files(
            quickstart_compose_file_name,
            quickstart_compose_file,
            graph_service_impl,
            kafka_setup,
            quickstart_arch,
            standalone_consumers,
            quickstart_execution_plan.composefile_git_ref,
        )

    # set version
    _set_environment_variables(
        version=quickstart_execution_plan.docker_tag,
        mysql_version=quickstart_execution_plan.mysql_tag,
        mysql_port=mysql_port,
        zk_port=zk_port,
        kafka_broker_port=kafka_broker_port,
        schema_registry_port=schema_registry_port,
        elastic_port=elastic_port,
        kafka_setup=kafka_setup,
    )

    compose = _docker_compose_v2()
    base_command: List[str] = [
        *compose,
        *itertools.chain.from_iterable(
            ("-f", f"{path}") for path in quickstart_compose_file
        ),
        "-p",
        DOCKER_COMPOSE_PROJECT_NAME,
    ]

    # Pull and possibly build the latest containers.
    try:
        if pull_images:
            click.echo("\nPulling docker images... ")
            click.secho(
                "This may take a while depending on your network bandwidth.", dim=True
            )

            # docker compose v2 seems to spam the stderr when used in a non-interactive environment.
            # As such, we'll only use the quiet flag if we're in an interactive environment.
            # If we're in quiet mode, then we'll show a spinner instead.
            quiet = not sys.stderr.isatty()
            with PerfTimer() as timer, click_spinner.spinner(disable=not quiet):
                subprocess.run(
                    [*base_command, "pull", *(("-q",) if quiet else ())],
                    check=True,
                    env=_docker_subprocess_env(),
                )

            telemetry.telemetry_instance.ping(
                "quickstart-image-pull",
                {
                    "status": "success",
                    "duration": timer.elapsed_seconds(),
                },
            )
            click.secho("Finished pulling docker images!")
    except subprocess.CalledProcessError:
        telemetry.telemetry_instance.ping(
            "quickstart-image-pull",
            {
                "status": "failure",
            },
        )
        click.secho(
            "Error while pulling images. Going to attempt to move on to docker compose up assuming the images have "
            "been built locally",
            fg="red",
        )

    if build_locally:
        logger.info("Building docker images locally...")
        subprocess.run(
            base_command + ["build", "--pull", "-q"],
            check=True,
            env=_docker_subprocess_env(),
        )
        logger.info("Finished building docker images!")

    # Start it up! (with retries)
    click.echo("\nStarting up DataHub...")
    start_time = datetime.datetime.now()
    status: Optional[QuickstartStatus] = None
    up_attempts = 0
    while (datetime.datetime.now() - start_time) < _QUICKSTART_MAX_WAIT_TIME:
        # We must run docker-compose up at least once.
        # Beyond that, we should run it again if something goes wrong.
        if up_attempts == 0 or (status and status.needs_up()):
            if up_attempts > 0:
                click.echo()
            up_attempts += 1

            logger.debug(f"Executing docker compose up command, attempt #{up_attempts}")
            up_process = subprocess.Popen(
                base_command + ["up", "-d", "--remove-orphans"],
                env=_docker_subprocess_env(),
            )
            try:
                up_process.wait(timeout=_QUICKSTART_UP_TIMEOUT.total_seconds())
            except subprocess.TimeoutExpired:
                logger.debug("docker compose up timed out, sending SIGTERM")
                up_process.terminate()
                try:
                    up_process.wait(timeout=8)
                except subprocess.TimeoutExpired:
                    logger.debug("docker compose up still running, sending SIGKILL")
                    up_process.kill()
                    up_process.wait()
            else:
                # If the docker process got a keyboard interrupt, raise one here.
                if up_process.returncode in {128 + signal.SIGINT, -signal.SIGINT}:
                    raise KeyboardInterrupt

        # Check docker health every few seconds.
        status = check_docker_quickstart()
        if status.is_ok():
            break

        # Wait until next iteration.
        click.echo(".", nl=False)
        time.sleep(_QUICKSTART_STATUS_CHECK_INTERVAL.total_seconds())
    else:
        assert status

        # Falls through if the while loop doesn't exit via break.
        click.echo()
        with tempfile.NamedTemporaryFile(suffix=".log", delete=False) as log_file:
            ret = subprocess.run(
                base_command + ["logs"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
                env=_docker_subprocess_env(),
            )
            log_file.write(ret.stdout)

        if dump_logs_on_failure:
            click.echo("Dumping docker compose logs:")
            click.echo(pathlib.Path(log_file.name).read_text())
            click.echo()

        raise status.to_exception(
            header="Unable to run quickstart - the following issues were detected:",
            footer="If you think something went wrong, please file an issue at https://github.com/datahub-project/datahub/issues\n"
            "or send a message in our Slack https://datahub.com/slack/\n"
            f"Be sure to attach the logs from {log_file.name}",
        )

    # Handle success condition.
    click.echo()
    click.secho("✔ DataHub is now running", fg="green")
    click.secho(
        "Ingest some demo data using `datahub docker ingest-sample-data`,\n"
        "or head to http://localhost:9002 (username: datahub, password: datahub) to play around with the frontend.",
        fg="green",
    )
    click.secho(
        "Need support? Get in touch on Slack: https://datahub.com/slack/",
        fg="magenta",
    )


def get_docker_compose_base_url(version_tag: str) -> str:
    if os.environ.get("DOCKER_COMPOSE_BASE"):
        return os.environ["DOCKER_COMPOSE_BASE"]

    return f"https://raw.githubusercontent.com/datahub-project/datahub/{version_tag}"


def get_github_file_url(neo4j: bool, is_m1: bool, release_version_tag: str) -> str:
    base_url = get_docker_compose_base_url(release_version_tag)
    if neo4j:
        github_file = (
            f"{base_url}/{NEO4J_AND_ELASTIC_QUICKSTART_COMPOSE_FILE}"
            if not is_m1
            else f"{base_url}/{NEO4J_AND_ELASTIC_M1_QUICKSTART_COMPOSE_FILE}"
        )
    else:
        github_file = (
            f"{base_url}/{ELASTIC_QUICKSTART_COMPOSE_FILE}"
            if not is_m1
            else f"{base_url}/{ELASTIC_M1_QUICKSTART_COMPOSE_FILE}"
        )
    return github_file


def download_compose_files(
    quickstart_compose_file_name,
    quickstart_compose_file_list,
    graph_service_impl,
    kafka_setup,
    quickstart_arch,
    standalone_consumers,
    compose_git_ref,
):
    # download appropriate quickstart file
    should_use_neo4j = should_use_neo4j_for_graph_service(graph_service_impl)
    is_m1 = is_arch_m1(quickstart_arch)
    github_file = get_github_file_url(should_use_neo4j, is_m1, compose_git_ref)
    # also allow local files
    request_session = requests.Session()
    request_session.mount("file://", FileAdapter())
    with (
        open(quickstart_compose_file_name, "wb")
        if quickstart_compose_file_name
        else tempfile.NamedTemporaryFile(suffix=".yml", delete=False)
    ) as tmp_file:
        path = pathlib.Path(tmp_file.name)
        quickstart_compose_file_list.append(path)
        logger.info(f"Fetching docker-compose file {github_file} from GitHub")
        # Download the quickstart docker-compose file from GitHub.
        quickstart_download_response = request_session.get(github_file)
        quickstart_download_response.raise_for_status()
        tmp_file.write(quickstart_download_response.content)
        logger.debug(f"Copied to {path}")
    if standalone_consumers:
        base_url = get_docker_compose_base_url(compose_git_ref)
        consumer_github_file = (
            f"{base_url}/{CONSUMERS_QUICKSTART_COMPOSE_FILE}"
            if should_use_neo4j
            else f"{base_url}/{ELASTIC_CONSUMERS_QUICKSTART_COMPOSE_FILE}"
        )

        default_consumer_compose_file = (
            Path(DATAHUB_ROOT_FOLDER) / "quickstart/docker-compose.consumers.yml"
        )
        with (
            open(default_consumer_compose_file, "wb")
            if default_consumer_compose_file
            else tempfile.NamedTemporaryFile(suffix=".yml", delete=False)
        ) as tmp_file:
            path = pathlib.Path(tmp_file.name)
            quickstart_compose_file_list.append(path)
            click.echo(
                f"Fetching consumer docker-compose file {consumer_github_file} from GitHub"
            )
            # Download the quickstart docker-compose file from GitHub.
            quickstart_download_response = request_session.get(consumer_github_file)
            quickstart_download_response.raise_for_status()
            tmp_file.write(quickstart_download_response.content)
            logger.debug(f"Copied to {path}")
    if kafka_setup:
        base_url = get_docker_compose_base_url(compose_git_ref)
        kafka_setup_github_file = f"{base_url}/{KAFKA_SETUP_QUICKSTART_COMPOSE_FILE}"

        default_kafka_compose_file = (
            Path(DATAHUB_ROOT_FOLDER) / "quickstart/docker-compose.kafka-setup.yml"
        )
        with (
            open(default_kafka_compose_file, "wb")
            if default_kafka_compose_file
            else tempfile.NamedTemporaryFile(suffix=".yml", delete=False)
        ) as tmp_file:
            path = pathlib.Path(tmp_file.name)
            quickstart_compose_file_list.append(path)
            click.echo(
                f"Fetching consumer docker-compose file {kafka_setup_github_file} from GitHub"
            )
            # Download the quickstart docker-compose file from GitHub.
            quickstart_download_response = request_session.get(kafka_setup_github_file)
            quickstart_download_response.raise_for_status()
            tmp_file.write(quickstart_download_response.content)
            logger.debug(f"Copied to {path}")


def valid_restore_options(
    restore: bool, restore_indices: bool, no_restore_indices: bool
) -> bool:
    if no_restore_indices and not restore:
        click.secho(
            "Using --no-restore-indices without a --restore isn't defined", fg="red"
        )
        return False
    if no_restore_indices and restore_indices:
        click.secho(
            "Using --restore-indices in conjunction with --no-restore-indices is undefined",
            fg="red",
        )
        return False
    if restore and restore_indices:
        click.secho(
            "Using --restore automatically implies using --restore-indices, you don't need to set both. Continuing...",
            fg="yellow",
        )
        return True
    return True


@docker.command()
@click.option(
    "--token",
    type=str,
    is_flag=False,
    default=None,
    help="The token to be used when ingesting, used when datahub is deployed with METADATA_SERVICE_AUTH_ENABLED=true",
)
@telemetry.with_telemetry()
def ingest_sample_data(token: Optional[str]) -> None:
    """Ingest sample data into a running DataHub instance."""

    # Verify that docker is up.
    status = check_docker_quickstart()
    if not status.is_ok():
        raise status.to_exception(
            header="Docker is not ready:",
            footer="Try running `datahub docker quickstart` first.",
        )

    # Run ingestion.
    click.echo("Starting ingestion...")
    recipe: dict = {
        "source": {
            "type": "demo-data",
            "config": {},
        },
        "sink": {
            "type": "datahub-rest",
            "config": {"server": "http://localhost:8080"},
        },
    }

    if token is not None:
        recipe["sink"]["config"]["token"] = token

    pipeline = Pipeline.create(recipe)
    pipeline.run()
    ret = pipeline.pretty_print_summary()
    sys.exit(ret)


@docker.command()
@telemetry.with_telemetry()
@click.option(
    "--keep-data",
    type=bool,
    is_flag=True,
    default=False,
    help="Delete data volumes",
)
def nuke(keep_data: bool) -> None:
    """Remove all Docker containers, networks, and volumes associated with DataHub."""

    with get_docker_client() as client:
        click.echo(f"Removing containers in the {DOCKER_COMPOSE_PROJECT_NAME} project")
        for container in client.containers.list(
            all=True, filters=DATAHUB_COMPOSE_PROJECT_FILTER
        ):
            container.remove(v=True, force=True)

        if keep_data:
            click.echo(
                f"Skipping deleting data volumes in the {DOCKER_COMPOSE_PROJECT_NAME} project"
            )
        else:
            click.echo(f"Removing volumes in the {DOCKER_COMPOSE_PROJECT_NAME} project")
            for filter in DATAHUB_COMPOSE_LEGACY_VOLUME_FILTERS + [
                DATAHUB_COMPOSE_PROJECT_FILTER
            ]:
                for volume in client.volumes.list(filters=filter):
                    volume.remove(force=True)

        click.echo(f"Removing networks in the {DOCKER_COMPOSE_PROJECT_NAME} project")
        for network in client.networks.list(filters=DATAHUB_COMPOSE_PROJECT_FILTER):
            network.remove()



import logging
import re
from dataclasses import dataclass, field
from enum import Enum
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
    Union,
)

from datahub.telemetry.telemetry import suppress_telemetry

logger = logging.getLogger(__name__)

# Only to be written to for logging server related information
global_debug: Dict[str, Any] = {}


def get_gms_config() -> Dict:
    return global_debug.get("gms_config", {})


class ServiceFeature(Enum):
    """
    Enum representing supported features in the REST service.
    """

    OPEN_API_SDK = "openapi_sdk"
    API_TRACING = "api_tracing"
    NO_CODE = "no_code"
    STATEFUL_INGESTION = "stateful_ingestion"
    IMPACT_ANALYSIS = "impact_analysis"
    PATCH_CAPABLE = "patch_capable"
    CLI_TELEMETRY = "cli_telemetry"
    DATAHUB_CLOUD = "datahub_cloud"
    # Add more features as needed


_REQUIRED_VERSION_OPENAPI_TRACING = {
    "cloud": (0, 3, 11, 0),
    "core": (1, 0, 1, 0),
}


@dataclass
class RestServiceConfig:
    """
    A class to represent REST service configuration with semantic version parsing capabilities.
    """

    raw_config: Dict[str, Any] = field(default_factory=dict)
    _version_cache: Optional[Tuple[int, int, int, int]] = None

    @property
    def commit_hash(self) -> Optional[str]:
        """
        Get the commit hash for the current version.

        Returns:
            The commit hash or None if not found
        """
        versions = self.raw_config.get("versions") or {}
        datahub_info = versions.get("acryldata/datahub") or {}
        return datahub_info.get("commit")

    @property
    def server_type(self) -> str:
        """
        Get the server type.

        Returns:
            The server type or "unknown" if not found
        """
        datahub = self.raw_config.get("datahub") or {}
        return datahub.get("serverType", "unknown")

    @property
    def service_version(self) -> Optional[str]:
        """
        Get the raw service version string.

        Returns:
            The version string or None if not found
        """
        versions = self.raw_config.get("versions") or {}
        datahub_info = versions.get("acryldata/datahub") or {}
        return datahub_info.get("version")

    def _parse_version(
        self, version_str: Optional[str] = None
    ) -> Tuple[int, int, int, int]:
        """
        Parse a semantic version string into its components, ignoring rc and suffixes.
        Supports standard three-part versions (1.0.0) and four-part versions (1.0.0.1).

        Args:
            version_str: Version string to parse. If None, uses the service version.

        Returns:
            Tuple of (major, minor, patch, build) version numbers where build is 0 for three-part versions

        Raises:
            ValueError: If the version string cannot be parsed
        """
        if version_str is None:
            version_str = self.service_version

        if not version_str:
            return (0, 0, 0, 0)

        # Remove 'v' prefix if present
        if version_str.startswith("v"):
            version_str = version_str[1:]

        # Extract the semantic version part (before any rc or suffix)
        # This pattern will match both three-part (1.0.0) and four-part (1.0.0.1) versions
        match = re.match(r"(\d+)\.(\d+)\.(\d+)(?:\.(\d+))?(?:rc\d+|-.*)?", version_str)
        if not match:
            raise ValueError(f"Invalid version format: {version_str}")

        major = int(match.group(1))
        minor = int(match.group(2))
        patch = int(match.group(3))
        build = (
            int(match.group(4)) if match.group(4) else 0
        )  # Default to 0 if not present

        return (major, minor, patch, build)

    @property
    def parsed_version(self) -> Optional[Tuple[int, int, int, int]]:
        """
        Get the parsed semantic version of the service.
        Uses caching for efficiency.

        Returns:
            Tuple of (major, minor, patch) version numbers
        """
        if self._version_cache is None:
            self._version_cache = self._parse_version()
        return self._version_cache

    def is_version_at_least(
        self, major: int, minor: int = 0, patch: int = 0, build: int = 0
    ) -> bool:
        """
        Check if the service version is at least the specified version.

        Args:
            major: Major version to check against
            minor: Minor version to check against
            patch: Patch version to check against
            build: Build version to check against (for four-part versions)

        Returns:
            True if the service version is at least the specified version
        """
        current_version = self.parsed_version or (0, 0, 0, 0)
        requested_version = (major, minor, patch, build)

        return current_version >= requested_version

    @property
    def is_no_code_enabled(self) -> bool:
        """
        Check if noCode is enabled.

        Returns:
            True if noCode is set to "true"
        """
        return self.raw_config.get("noCode") == "true"

    @property
    def is_managed_ingestion_enabled(self) -> bool:
        """
        Check if managedIngestion is enabled.

        Returns:
            True if managedIngestion.enabled is True
        """
        managed_ingestion = self.raw_config.get("managedIngestion") or {}
        return managed_ingestion.get("enabled", False)

    @property
    def is_datahub_cloud(self) -> bool:
        """
        Check if DataHub Cloud is enabled.

        Returns:
            True if the server environment is not 'core'
        """
        datahub_config = self.raw_config.get("datahub") or {}
        server_env = datahub_config.get("serverEnv")

        # Return False if serverEnv is None or empty string
        if not server_env:
            return False

        return server_env != "core"

    def supports_feature(self, feature: ServiceFeature) -> bool:
        """
        Determines whether a specific feature is supported based on service version
        and whether this is a cloud deployment or not.

        Args:
            feature: Feature enum value to check

        Returns:
            Boolean indicating whether the feature is supported
        """
        # Special handling for features that rely on config flags
        config_based_features = {
            ServiceFeature.NO_CODE: lambda: self.is_no_code_enabled,
            ServiceFeature.STATEFUL_INGESTION: lambda: self.raw_config.get(
                "statefulIngestionCapable", False
            )
            is True,
            ServiceFeature.IMPACT_ANALYSIS: lambda: self.raw_config.get(
                "supportsImpactAnalysis", False
            )
            is True,
            ServiceFeature.PATCH_CAPABLE: lambda: self.raw_config.get(
                "patchCapable", False
            )
            is True,
            ServiceFeature.CLI_TELEMETRY: lambda: (
                self.raw_config.get("telemetry") or {}
            ).get("enabledCli", None),
            ServiceFeature.DATAHUB_CLOUD: lambda: self.is_datahub_cloud,
        }

        # Check if this is a config-based feature
        if feature in config_based_features:
            return config_based_features[feature]()

        # For environment-based features, determine requirements based on cloud vs. non-cloud
        deployment_type = "cloud" if self.is_datahub_cloud else "core"

        # Define feature requirements
        feature_requirements = {
            ServiceFeature.OPEN_API_SDK: _REQUIRED_VERSION_OPENAPI_TRACING,
            ServiceFeature.API_TRACING: _REQUIRED_VERSION_OPENAPI_TRACING,
            # Additional features can be defined here
        }

        # Check if the feature exists in our requirements dictionary
        if feature not in feature_requirements:
            # Unknown feature, assume not supported
            return False

        # Get version requirements for this feature and deployment type
        feature_reqs = feature_requirements[feature]
        requirements = feature_reqs.get(deployment_type)

        if not requirements:
            # If no specific requirements defined for this deployment type,
            # assume feature is not supported
            return False

        # Check if the current version meets the requirements
        req_major, req_minor, req_patch, req_build = requirements
        return self.is_version_at_least(req_major, req_minor, req_patch, req_build)

    def __str__(self) -> str:
        """
        Return a string representation of the configuration as JSON.

        Returns:
            A string representation of the configuration dictionary
        """
        return str(self.raw_config)

    def __repr__(self) -> str:
        """
        Return a representation of the object that can be used to recreate it.

        Returns:
            A string representation that can be used with pprint
        """
        return str(self.raw_config)


def set_gms_config(config: Union[Dict[str, Any], RestServiceConfig]) -> None:
    global_debug["gms_config"] = config

    config_obj = (
        config
        if isinstance(config, RestServiceConfig)
        else RestServiceConfig(raw_config=config)
    )

    cli_telemetry_enabled = is_cli_telemetry_enabled(config_obj)
    if cli_telemetry_enabled is not None and not cli_telemetry_enabled:
        # server requires telemetry to be disabled on client
        suppress_telemetry()


def is_cli_telemetry_enabled(config: RestServiceConfig) -> bool:
    return config.supports_feature(ServiceFeature.CLI_TELEMETRY)



from typing import Dict, Type

from datahub.metadata.schema_classes import ASPECT_CLASSES, _Aspect

ASPECT_MAP: Dict[str, Type[_Aspect]] = {
    AspectClass.get_aspect_name(): AspectClass for AspectClass in ASPECT_CLASSES
}

TIMESERIES_ASPECT_MAP: Dict[str, Type[_Aspect]] = {
    name: klass
    for name, klass in ASPECT_MAP.items()
    if klass.get_aspect_type() == "timeseries"
}

JSON_CONTENT_TYPE = "application/json"
JSON_PATCH_CONTENT_TYPE = "application/json-patch+json"



import logging
import uuid
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import yaml
from pydantic import validator
from ruamel.yaml import YAML
from typing_extensions import Literal

from datahub.api.entities.forms.forms_graphql_constants import (
    CREATE_DYNAMIC_FORM_ASSIGNMENT,
    FIELD_FILTER_TEMPLATE,
    UPLOAD_ENTITIES_FOR_FORMS,
)
from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import (
    make_container_urn,
    make_data_platform_urn,
    make_domain_urn,
    make_group_urn,
    make_tag_urn,
    make_term_urn,
    make_user_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.graph.client import DataHubGraph, get_default_graph
from datahub.ingestion.graph.config import ClientMode
from datahub.metadata.schema_classes import (
    FormActorAssignmentClass,
    FormInfoClass,
    FormPromptClass,
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
    StructuredPropertyParamsClass,
)
from datahub.utilities.urns.urn import Urn

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

FILTER_CRITERION_TYPES = "_entityType"
FILTER_CRITERION_SUB_TYPES = "typeNames.keyword"
FILTER_CRITERION_PLATFORMS = "platform.keyword"
FILTER_CRITERION_PLATFORM_INSTANCES = "dataPlatformInstance.keyword"
FILTER_CRITERION_DOMAINS = "domains.keyword"
FILTER_CRITERION_CONTAINERS = "container.keyword"
FILTER_CRITERION_OWNERS = "owners.keyword"
FILTER_CRITERION_TAGS = "tags.keyword"
FILTER_CRITERION_GLOSSARY_TERMS = "glossaryTerms.keyword"


class PromptType(Enum):
    STRUCTURED_PROPERTY = "STRUCTURED_PROPERTY"
    FIELDS_STRUCTURED_PROPERTY = "FIELDS_STRUCTURED_PROPERTY"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class Prompt(ConfigModel):
    id: Optional[str] = None
    title: str
    description: Optional[str] = None
    type: str
    structured_property_id: Optional[str] = None
    structured_property_urn: Optional[str] = None
    required: Optional[bool] = None

    @validator("structured_property_urn", pre=True, always=True)
    def structured_property_urn_must_be_present(cls, v, values):
        if not v and values.get("structured_property_id"):
            return Urn.make_structured_property_urn(values["structured_property_id"])
        return v


class FormType(Enum):
    VERIFICATION = "VERIFICATION"
    DOCUMENTATION = "COMPLETION"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class Filters(ConfigModel):
    types: Optional[List[str]] = None
    sub_types: Optional[List[str]] = None
    platforms: Optional[List[str]] = None
    platform_instances: Optional[List[str]] = None
    domains: Optional[List[str]] = None
    containers: Optional[List[str]] = None
    owners: Optional[List[str]] = None
    tags: Optional[List[str]] = None
    terms: Optional[List[str]] = None


class Entities(ConfigModel):
    urns: Optional[List[str]] = None
    filters: Optional[Filters] = None


class Actors(ConfigModel):
    owners: Optional[bool] = None
    users: Optional[List[str]] = None  # can be user IDs or urns
    groups: Optional[List[str]] = None  # can be group IDs or urns


class Forms(ConfigModel):
    id: Optional[str] = None
    urn: Optional[str] = None
    name: str
    description: Optional[str] = None
    prompts: List[Prompt] = []
    type: Optional[str] = None
    version: Optional[Literal[1]] = None
    entities: Optional[Entities] = None
    owners: Optional[List[str]] = None  # can be user IDs or urns
    group_owners: Optional[List[str]] = None  # can be group IDs or urns
    actors: Optional[Actors] = None

    @validator("urn", pre=True, always=True)
    def urn_must_be_present(cls, v, values):
        if not v:
            if values.get("id") is None:
                raise ValueError("Form id must be present if urn is not")
            return f"urn:li:form:{values['id']}"
        return v

    @staticmethod
    def create(file: str) -> None:
        emitter: DataHubGraph

        with get_default_graph(ClientMode.CLI) as emitter, open(file) as fp:
            forms: List[dict] = yaml.safe_load(fp)
            for form_raw in forms:
                form = Forms.parse_obj(form_raw)

                try:
                    if not FormType.has_value(form.type):
                        logger.error(
                            f"Form type {form.type} does not exist. Please try again with a valid type."
                        )

                    mcp = MetadataChangeProposalWrapper(
                        entityUrn=form.urn,
                        aspect=FormInfoClass(
                            name=form.name,
                            description=form.description,
                            prompts=form.validate_prompts(emitter),
                            type=form.type,
                            actors=form.create_form_actors(form.actors),
                        ),
                    )
                    emitter.emit_mcp(mcp)

                    logger.info(f"Created form {form.urn}")

                    if form.owners or form.group_owners:
                        form.add_owners(emitter)

                    if form.entities:
                        if form.entities.urns:
                            # Associate specific entities with a form
                            form.upload_entities_for_form(emitter)

                        if form.entities.filters:
                            # Associate groups of entities with a form based on filters
                            form.create_form_filters(emitter)

                except Exception as e:
                    logger.error(e)
                    return

    def validate_prompts(self, emitter: DataHubGraph) -> List[FormPromptClass]:
        prompts = []
        if self.prompts:
            for prompt in self.prompts:
                if not prompt.id:
                    prompt.id = str(uuid.uuid4())
                    logger.warning(
                        f"Prompt id not provided. Setting prompt id to {prompt.id}"
                    )
                if prompt.structured_property_urn:
                    structured_property_urn = prompt.structured_property_urn
                    if emitter.exists(structured_property_urn):
                        prompt.structured_property_urn = structured_property_urn
                    else:
                        raise Exception(
                            f"Structured property {structured_property_urn} does not exist. Unable to create form."
                        )
                elif (
                    prompt.type
                    in (
                        PromptType.STRUCTURED_PROPERTY.value,
                        PromptType.FIELDS_STRUCTURED_PROPERTY.value,
                    )
                    and not prompt.structured_property_urn
                ):
                    raise Exception(
                        f"Prompt type is {prompt.type} but no structured properties exist. Unable to create form."
                    )
                if (
                    prompt.type == PromptType.FIELDS_STRUCTURED_PROPERTY.value
                    and prompt.required
                ):
                    raise Exception(
                        "Schema field prompts cannot be marked as required. Ensure these prompts are not required."
                    )

                prompts.append(
                    FormPromptClass(
                        id=prompt.id,
                        title=prompt.title,
                        description=prompt.description,
                        type=prompt.type,
                        structuredPropertyParams=(
                            StructuredPropertyParamsClass(
                                urn=prompt.structured_property_urn
                            )
                            if prompt.structured_property_urn
                            else None
                        ),
                        required=prompt.required,
                    )
                )
        else:
            logger.warning(f"No prompts exist on form {self.urn}. Is that intended?")

        return prompts

    def create_form_actors(
        self, actors: Optional[Actors] = None
    ) -> Union[None, FormActorAssignmentClass]:
        if actors is None:
            return None

        users = None
        if actors.users is not None:
            users = Forms.format_users(actors.users)

        groups = None
        if actors.groups is not None:
            groups = Forms.format_groups(actors.groups)

        return FormActorAssignmentClass(
            owners=actors.owners, users=users, groups=groups
        )

    def upload_entities_for_form(self, emitter: DataHubGraph) -> Union[None, Exception]:
        if self.entities and self.entities.urns:
            formatted_entity_urns = ", ".join(
                [f'"{value}"' for value in self.entities.urns]
            )
            query = UPLOAD_ENTITIES_FOR_FORMS.format(
                form_urn=self.urn, entity_urns=formatted_entity_urns
            )
            result = emitter.execute_graphql(query=query)
            if not result:
                return Exception(f"Could not bulk upload entities for form {self.urn}.")

        return None

    def create_form_filters(self, emitter: DataHubGraph) -> Union[None, Exception]:
        filters_raw = []
        # Loop through each entity and assign a filter for it
        if self.entities and self.entities.filters:
            filters = self.entities.filters

            if filters.types:
                filters_raw.append(
                    Forms.format_form_filter(FILTER_CRITERION_TYPES, filters.types)
                )

            if filters.sub_types:
                filters_raw.append(
                    Forms.format_form_filter(
                        FILTER_CRITERION_SUB_TYPES, filters.sub_types
                    )
                )

            if filters.platforms:
                urns = [
                    make_data_platform_urn(platform) for platform in filters.platforms
                ]
                filters_raw.append(
                    Forms.format_form_filter(FILTER_CRITERION_PLATFORMS, urns)
                )

            if filters.platform_instances:
                urns = []
                for platform_instance in filters.platform_instances:
                    platform_instance_urn = Forms.validate_platform_instance_urn(
                        platform_instance
                    )
                    if platform_instance_urn:
                        urns.append(platform_instance_urn)
                filters_raw.append(
                    Forms.format_form_filter(FILTER_CRITERION_PLATFORM_INSTANCES, urns)
                )

            if filters.domains:
                urns = [make_domain_urn(domain) for domain in filters.domains]
                filters_raw.append(
                    Forms.format_form_filter(FILTER_CRITERION_DOMAINS, urns)
                )

            if filters.containers:
                urns = [
                    make_container_urn(container) for container in filters.containers
                ]
                filters_raw.append(
                    Forms.format_form_filter(FILTER_CRITERION_CONTAINERS, urns)
                )

            if filters.owners:
                urns = [make_user_urn(owner) for owner in filters.owners]
                filters_raw.append(
                    Forms.format_form_filter(FILTER_CRITERION_OWNERS, urns)
                )

            if filters.tags:
                urns = [make_tag_urn(tag) for tag in filters.tags]
                filters_raw.append(
                    Forms.format_form_filter(FILTER_CRITERION_TAGS, urns)
                )

            if filters.terms:
                urns = [make_term_urn(term) for term in filters.terms]
                filters_raw.append(
                    Forms.format_form_filter(FILTER_CRITERION_GLOSSARY_TERMS, urns)
                )

        filters_str = ", ".join(item for item in filters_raw)
        result = emitter.execute_graphql(
            query=CREATE_DYNAMIC_FORM_ASSIGNMENT.format(
                form_urn=self.urn, filters=filters_str
            )
        )

        if not result:
            return Exception(
                f"Could not bulk upload urns or filters for form {self.urn}."
            )

        return None

    def add_owners(self, emitter: DataHubGraph) -> Union[None, Exception]:
        owner_urns: List[str] = []
        if self.owners:
            owner_urns += Forms.format_users(self.owners)
        if self.group_owners:
            owner_urns += Forms.format_groups(self.group_owners)

        ownership = OwnershipClass(
            owners=[
                OwnerClass(owner=urn, type=OwnershipTypeClass.TECHNICAL_OWNER)
                for urn in (owner_urns or [])
            ],
        )

        try:
            mcp = MetadataChangeProposalWrapper(entityUrn=self.urn, aspect=ownership)
            emitter.emit_mcp(mcp)
        except Exception as e:
            logger.error(e)

        return None

    @staticmethod
    def format_form_filter(field: str, urns: List[str]) -> str:
        formatted_urns = ", ".join([f'"{urn}"' for urn in urns])
        return FIELD_FILTER_TEMPLATE.format(field=field, values=formatted_urns)

    @staticmethod
    def validate_platform_instance_urn(instance: str) -> Union[str, None]:
        if instance.startswith("urn:li:dataPlatformInstance:"):
            return instance

        logger.warning(
            f"{instance} is not an urn. Unable to create platform instance filter."
        )
        return None

    @staticmethod
    def from_datahub(graph: DataHubGraph, urn: str) -> "Forms":
        form: Optional[FormInfoClass] = graph.get_aspect(urn, FormInfoClass)
        if form is None:
            raise Exception("FormInfo aspect is None. Unable to create form.")
        prompts = []
        for prompt_raw in form.prompts:
            prompts.append(
                Prompt(
                    id=prompt_raw.id,
                    title=prompt_raw.title,
                    description=prompt_raw.description,
                    type=prompt_raw.type,
                    structured_property_urn=(
                        prompt_raw.structuredPropertyParams.urn
                        if prompt_raw.structuredPropertyParams
                        else None
                    ),
                )
            )
        return Forms(
            urn=urn,
            name=form.name,
            description=form.description,
            prompts=prompts,
            type=form.type,
        )

    @staticmethod
    def format_users(users: List[str]) -> List[str]:
        formatted_users: List[str] = []

        for user in users:
            if user.startswith("urn:li:"):
                formatted_users.append(user)
            else:
                formatted_users.append(make_user_urn(user))

        return formatted_users

    @staticmethod
    def format_groups(groups: List[str]) -> List[str]:
        formatted_groups: List[str] = []

        for group in groups:
            if group.startswith("urn:li:"):
                formatted_groups.append(group)
            else:
                formatted_groups.append(make_group_urn(group))

        return formatted_groups

    def to_yaml(
        self,
        file: Path,
    ) -> None:
        with open(file, "w") as fp:
            yaml = YAML(typ="rt")  # default, if not specfied, is 'rt' (round-trip)
            yaml.indent(mapping=2, sequence=4, offset=2)
            yaml.default_flow_style = False
            yaml.dump(self.dict(), fp)



import logging
from typing import Iterable, Optional

from pydantic.fields import Field

from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import datahub_guid, set_aspect
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.metadata.schema_classes import (
    ChartInfoClass,
    DashboardInfoClass,
    FineGrainedLineageClass,
    MetadataChangeEventClass,
    SystemMetadataClass,
    UpstreamLineageClass,
)
from datahub.specific.chart import ChartPatchBuilder
from datahub.specific.dashboard import DashboardPatchBuilder
from datahub.specific.dataset import DatasetPatchBuilder

logger = logging.getLogger(__name__)


def convert_upstream_lineage_to_patch(
    urn: str,
    aspect: UpstreamLineageClass,
    system_metadata: Optional[SystemMetadataClass],
) -> MetadataWorkUnit:
    patch_builder = DatasetPatchBuilder(urn, system_metadata)
    for upstream in aspect.upstreams:
        patch_builder.add_upstream_lineage(upstream)
    for fine_upstream in aspect.fineGrainedLineages or []:
        patch_builder.add_fine_grained_upstream_lineage(fine_upstream)
    mcp = next(iter(patch_builder.build()))
    return MetadataWorkUnit(id=MetadataWorkUnit.generate_workunit_id(mcp), mcp_raw=mcp)


def convert_chart_info_to_patch(
    urn: str, aspect: ChartInfoClass, system_metadata: Optional[SystemMetadataClass]
) -> Optional[MetadataWorkUnit]:
    patch_builder = ChartPatchBuilder(urn, system_metadata)

    if aspect.customProperties:
        for key in aspect.customProperties:
            patch_builder.add_custom_property(
                key, str(aspect.customProperties.get(key))
            )

    if aspect.inputEdges:
        for inputEdge in aspect.inputEdges:
            patch_builder.add_input_edge(inputEdge)

    patch_builder.set_chart_url(aspect.chartUrl).set_external_url(
        aspect.externalUrl
    ).set_type(aspect.type).set_title(aspect.title).set_access(
        aspect.access
    ).set_last_modified(aspect.lastModified).set_last_refreshed(
        aspect.lastRefreshed
    ).set_description(aspect.description).add_inputs(aspect.inputs)

    values = patch_builder.build()
    if values:
        mcp = next(iter(values))
        return MetadataWorkUnit(
            id=MetadataWorkUnit.generate_workunit_id(mcp), mcp_raw=mcp
        )
    return None


def convert_dashboard_info_to_patch(
    urn: str, aspect: DashboardInfoClass, system_metadata: Optional[SystemMetadataClass]
) -> Optional[MetadataWorkUnit]:
    patch_builder = DashboardPatchBuilder(urn, system_metadata)

    if aspect.customProperties:
        for key in aspect.customProperties:
            patch_builder.add_custom_property(
                key, str(aspect.customProperties.get(key))
            )

    if aspect.datasetEdges:
        for datasetEdge in aspect.datasetEdges:
            patch_builder.add_dataset_edge(datasetEdge)

    if aspect.chartEdges:
        for chartEdge in aspect.chartEdges:
            patch_builder.add_chart_edge(chartEdge)

    if aspect.title:
        patch_builder.set_title(aspect.title)

    if aspect.description:
        patch_builder.set_description(aspect.description)

    if aspect.charts:
        patch_builder.add_charts(aspect.charts)

    if aspect.dashboardUrl:
        patch_builder.set_dashboard_url(aspect.dashboardUrl)

    if aspect.datasets:
        patch_builder.add_datasets(aspect.datasets)

    if aspect.dashboards:
        for dashboard in aspect.dashboards:
            patch_builder.add_dashboard(dashboard)

    if aspect.access:
        patch_builder.set_access(aspect.access)

    if aspect.lastRefreshed:
        patch_builder.set_last_refreshed(aspect.lastRefreshed)

    if aspect.lastModified:
        patch_builder.set_last_modified(last_modified=aspect.lastModified)

    values = patch_builder.build()

    if values:
        logger.debug(
            f"Generating patch DashboardInfo MetadataWorkUnit for dashboard {aspect.title}"
        )
        mcp = next(iter(values))
        return MetadataWorkUnit(
            id=MetadataWorkUnit.generate_workunit_id(mcp), mcp_raw=mcp
        )
    return None


def get_fine_grained_lineage_key(fine_upstream: FineGrainedLineageClass) -> str:
    return datahub_guid(
        {
            "upstreams": sorted(fine_upstream.upstreams or []),
            "downstreams": sorted(fine_upstream.downstreams or []),
            "transformOperation": fine_upstream.transformOperation,
        }
    )


def auto_incremental_lineage(
    incremental_lineage: bool,
    stream: Iterable[MetadataWorkUnit],
) -> Iterable[MetadataWorkUnit]:
    if not incremental_lineage:
        yield from stream
        return  # early exit

    for wu in stream:
        urn = wu.get_urn()

        if isinstance(wu.metadata, MetadataChangeEventClass):
            lineage_aspect = wu.get_aspect_of_type(UpstreamLineageClass)
            set_aspect(wu.metadata, None, UpstreamLineageClass)
            if len(wu.metadata.proposedSnapshot.aspects) > 0:
                yield wu

            if lineage_aspect and lineage_aspect.upstreams:
                yield convert_upstream_lineage_to_patch(
                    urn, lineage_aspect, wu.metadata.systemMetadata
                )
        elif isinstance(wu.metadata, MetadataChangeProposalWrapper) and isinstance(
            wu.metadata.aspect, UpstreamLineageClass
        ):
            lineage_aspect = wu.metadata.aspect
            if lineage_aspect.upstreams:
                yield convert_upstream_lineage_to_patch(
                    urn, lineage_aspect, wu.metadata.systemMetadata
                )
        else:
            yield wu


class IncrementalLineageConfigMixin(ConfigModel):
    incremental_lineage: bool = Field(
        default=False,
        description="When enabled, emits lineage as incremental to existing lineage already in DataHub. When disabled, re-states lineage on each run.",
    )



from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Optional, Type, Union

from typing_extensions import Self

import datahub.metadata.schema_classes as models
from datahub.metadata.urns import ChartUrn, DashboardUrn, DatasetUrn, Urn
from datahub.sdk._shared import (
    ChartUrnOrStr,
    DashboardUrnOrStr,
    DataPlatformInstanceUrnOrStr,
    DataPlatformUrnOrStr,
    DatasetUrnOrStr,
    DomainInputType,
    HasContainer,
    HasDomain,
    HasInstitutionalMemory,
    HasOwnership,
    HasPlatformInstance,
    HasSubtype,
    HasTags,
    HasTerms,
    LinksInputType,
    OwnersInputType,
    TagsInputType,
    TermsInputType,
)
from datahub.sdk.chart import Chart
from datahub.sdk.dataset import Dataset
from datahub.sdk.entity import Entity, ExtraAspectsType


class Dashboard(
    HasPlatformInstance,
    HasSubtype,
    HasOwnership,
    HasContainer,
    HasInstitutionalMemory,
    HasTags,
    HasTerms,
    HasDomain,
    Entity,
):
    """Represents a dashboard in DataHub."""

    __slots__ = ()

    @classmethod
    def get_urn_type(cls) -> Type[DashboardUrn]:
        """Get the URN type for dashboards.
        Returns:
            The DashboardUrn class.
        """
        return DashboardUrn

    def __init__(
        self,
        *,
        # Identity.
        name: str,
        platform: DataPlatformUrnOrStr,
        display_name: Optional[str] = None,
        platform_instance: Optional[DataPlatformInstanceUrnOrStr] = None,
        # Dashboard properties.
        description: str = "",
        external_url: Optional[str] = None,
        dashboard_url: Optional[str] = None,
        custom_properties: Optional[Dict[str, str]] = None,
        last_modified: Optional[datetime] = None,
        last_refreshed: Optional[datetime] = None,
        input_datasets: Optional[List[Union[DatasetUrnOrStr, Dataset]]] = None,
        charts: Optional[List[Union[ChartUrnOrStr, Chart]]] = None,
        dashboards: Optional[List[Union[DashboardUrnOrStr, Dashboard]]] = None,
        # Standard aspects.
        subtype: Optional[str] = None,
        owners: Optional[OwnersInputType] = None,
        links: Optional[LinksInputType] = None,
        tags: Optional[TagsInputType] = None,
        terms: Optional[TermsInputType] = None,
        domain: Optional[DomainInputType] = None,
        extra_aspects: ExtraAspectsType = None,
    ):
        """Initialize a new Dashboard instance."""
        urn = DashboardUrn.create_from_ids(
            platform=str(platform),
            name=name,
            platform_instance=str(platform_instance) if platform_instance else None,
        )
        super().__init__(urn)
        self._set_extra_aspects(extra_aspects)

        self._set_platform_instance(platform, platform_instance)

        # Initialize DashboardInfoClass with default values
        dashboard_info = models.DashboardInfoClass(
            title=display_name or name,
            description=description or "",
            lastModified=models.ChangeAuditStampsClass(
                lastModified=None,
            ),
            customProperties={},
            chartEdges=[],
            datasetEdges=[],
            dashboards=[],
        )

        if last_modified:
            dashboard_info.lastModified = models.ChangeAuditStampsClass(
                lastModified=models.AuditStampClass(
                    time=int(last_modified.timestamp()),
                    actor="urn:li:corpuser:datahub",
                ),
            )

        # Set additional properties
        if description is not None:
            self.set_description(description)
        if display_name is not None:
            self.set_display_name(display_name)
        if external_url is not None:
            self.set_external_url(external_url)
        if dashboard_url is not None:
            self.set_dashboard_url(dashboard_url)
        if custom_properties is not None:
            self.set_custom_properties(custom_properties)
        if last_modified is not None:
            self.set_last_modified(last_modified)
        if last_refreshed is not None:
            self.set_last_refreshed(last_refreshed)
        if subtype is not None:
            self.set_subtype(subtype)
        if owners is not None:
            self.set_owners(owners)
        if links is not None:
            self.set_links(links)
        if tags is not None:
            self.set_tags(tags)
        if terms is not None:
            self.set_terms(terms)
        if domain is not None:
            self.set_domain(domain)
        if input_datasets is not None:
            self.set_input_datasets(input_datasets)
        if charts is not None:
            self.set_charts(charts)
        if dashboards is not None:
            self.set_dashboards(dashboards)

    @classmethod
    def _new_from_graph(cls, urn: Urn, current_aspects: models.AspectBag) -> Self:
        assert isinstance(urn, DashboardUrn)
        entity = cls(
            platform=urn.dashboard_tool,
            name=urn.dashboard_id,
        )
        return entity._init_from_graph(current_aspects)

    @property
    def urn(self) -> DashboardUrn:
        assert isinstance(self._urn, DashboardUrn)
        return self._urn

    def _ensure_dashboard_props(self) -> models.DashboardInfoClass:
        """Get the dashboard properties safely."""
        return self._setdefault_aspect(
            models.DashboardInfoClass(
                title=self.urn.dashboard_id,
                description="",
                lastModified=models.ChangeAuditStampsClass(
                    lastModified=models.AuditStampClass(
                        time=0, actor="urn:li:corpuser:unknown"
                    )
                ),
                customProperties={},
                chartEdges=[],
                datasetEdges=[],
                dashboards=[],
            )
        )

    @property
    def name(self) -> str:
        """Get the name of the dashboard."""
        return self.urn.dashboard_id

    @property
    def title(self) -> str:
        """Get the title of the dashboard."""
        return self._ensure_dashboard_props().title

    def set_title(self, title: str) -> None:
        """Set the title of the dashboard."""
        props = self._ensure_dashboard_props()
        props.title = title
        self._set_aspect(props)

    @property
    def description(self) -> Optional[str]:
        """Get the description of the dashboard."""
        props = self._ensure_dashboard_props()
        return props.description

    def set_description(self, description: str) -> None:
        """Set the description of the dashboard."""
        props = self._ensure_dashboard_props()
        props.description = description
        self._set_aspect(props)

    @property
    def display_name(self) -> Optional[str]:
        """Get the display name of the dashboard."""
        return self.title

    def set_display_name(self, display_name: str) -> None:
        """Set the display name of the dashboard."""
        self.set_title(display_name)

    @property
    def external_url(self) -> Optional[str]:
        """Get the external URL of the dashboard."""
        props = self._ensure_dashboard_props()
        return props.externalUrl

    def set_external_url(self, external_url: str) -> None:
        """Set the external URL of the dashboard."""
        props = self._ensure_dashboard_props()
        props.externalUrl = external_url
        self._set_aspect(props)

    @property
    def dashboard_url(self) -> Optional[str]:
        """Get the dashboard URL."""
        props = self._ensure_dashboard_props()
        return props.dashboardUrl

    def set_dashboard_url(self, dashboard_url: str) -> None:
        """Set the dashboard URL."""
        props = self._ensure_dashboard_props()
        props.dashboardUrl = dashboard_url
        self._set_aspect(props)

    @property
    def custom_properties(self) -> Dict[str, str]:
        """Get the custom properties of the dashboard."""
        props = self._ensure_dashboard_props()
        return props.customProperties or {}

    def set_custom_properties(self, custom_properties: Dict[str, str]) -> None:
        """Set the custom properties of the dashboard."""
        props = self._ensure_dashboard_props()
        props.customProperties = custom_properties
        self._set_aspect(props)

    @property
    def last_modified(self) -> Optional[datetime]:
        """Get the last modification timestamp of the dashboard."""
        props = self._ensure_dashboard_props()
        if props.lastModified.lastModified.time == 0:
            return None
        return datetime.fromtimestamp(props.lastModified.lastModified.time)

    def set_last_modified(self, last_modified: datetime) -> None:
        """Set the last modification timestamp of the dashboard."""
        props = self._ensure_dashboard_props()
        props.lastModified = models.ChangeAuditStampsClass(
            lastModified=models.AuditStampClass(
                time=int(last_modified.timestamp()),
                actor="urn:li:corpuser:datahub",
            ),
        )
        self._set_aspect(props)

    @property
    def last_refreshed(self) -> Optional[datetime]:
        """Get the last refresh timestamp of the dashboard."""
        props = self._ensure_dashboard_props()
        return (
            datetime.fromtimestamp(props.lastRefreshed)
            if props.lastRefreshed is not None
            else None
        )

    def set_last_refreshed(self, last_refreshed: datetime) -> None:
        """Set the last refresh timestamp of the dashboard."""
        props = self._ensure_dashboard_props()
        props.lastRefreshed = int(last_refreshed.timestamp())
        self._set_aspect(props)

    @property
    def input_datasets(self) -> List[DatasetUrn]:
        """Get the input datasets of the dashboard."""
        props = self._ensure_dashboard_props()
        return [
            DatasetUrn.from_string(edge.destinationUrn)
            for edge in (props.datasetEdges or [])
        ]

    def set_input_datasets(
        self, input_datasets: List[Union[DatasetUrnOrStr, Dataset]]
    ) -> None:
        """Set the input datasets of the dashboard."""
        props = self._ensure_dashboard_props()
        dataset_edges = props.datasetEdges or []
        for dataset in input_datasets:
            if isinstance(dataset, Dataset):
                dataset_urn = dataset.urn
            else:
                dataset_urn = DatasetUrn.from_string(dataset)
            dataset_edges.append(models.EdgeClass(destinationUrn=str(dataset_urn)))
        props.datasetEdges = dataset_edges
        self._set_aspect(props)

    def add_input_dataset(self, input_dataset: Union[DatasetUrnOrStr, Dataset]) -> None:
        """Add an input dataset to the dashboard."""
        if isinstance(input_dataset, Dataset):
            input_dataset_urn = input_dataset.urn
        else:
            input_dataset_urn = DatasetUrn.from_string(input_dataset)
        props = self._ensure_dashboard_props()
        dataset_edges = props.datasetEdges or []
        existing_urns = [edge.destinationUrn for edge in dataset_edges]
        if str(input_dataset_urn) not in existing_urns:
            dataset_edges.append(
                models.EdgeClass(destinationUrn=str(input_dataset_urn))
            )
        props.datasetEdges = dataset_edges
        self._set_aspect(props)

    def remove_input_dataset(
        self, input_dataset: Union[DatasetUrnOrStr, Dataset]
    ) -> None:
        """Remove an input dataset from the dashboard."""
        if isinstance(input_dataset, Dataset):
            input_dataset_urn = input_dataset.urn
        else:
            input_dataset_urn = DatasetUrn.from_string(input_dataset)
        props = self._ensure_dashboard_props()
        props.datasetEdges = [
            edge
            for edge in (props.datasetEdges or [])
            if edge.destinationUrn != str(input_dataset_urn)
        ]
        self._set_aspect(props)

    @property
    def charts(self) -> List[ChartUrn]:
        """Get the charts of the dashboard."""
        chart_edges = self._ensure_dashboard_props().chartEdges
        if chart_edges is None:
            return []
        return [ChartUrn.from_string(edge.destinationUrn) for edge in chart_edges]

    def set_charts(self, charts: List[Union[ChartUrnOrStr, Chart]]) -> None:
        """Set the charts of the dashboard."""
        props = self._ensure_dashboard_props()
        chart_edges = props.chartEdges or []
        for chart in charts:
            if isinstance(chart, Chart):
                chart_urn = chart.urn
            else:
                chart_urn = ChartUrn.from_string(chart)
            chart_edges.append(models.EdgeClass(destinationUrn=str(chart_urn)))
        props.chartEdges = chart_edges
        self._set_aspect(props)

    def add_chart(self, chart: Union[ChartUrnOrStr, Chart]) -> None:
        """Add a chart to the dashboard."""
        if isinstance(chart, Chart):
            chart_urn = chart.urn
        else:
            chart_urn = ChartUrn.from_string(chart)
        props = self._ensure_dashboard_props()
        chart_edges = props.chartEdges or []
        existing_urns = [
            edge.destinationUrn
            for edge in chart_edges
            if edge.destinationUrn is not None
        ]
        if str(chart_urn) not in existing_urns:
            chart_edges.append(models.EdgeClass(destinationUrn=str(chart_urn)))
        props.chartEdges = chart_edges
        self._set_aspect(props)

    def remove_chart(self, chart: Union[ChartUrnOrStr, Chart]) -> None:
        """Remove a chart from the dashboard."""
        if isinstance(chart, Chart):
            chart_urn = chart.urn
        else:
            chart_urn = ChartUrn.from_string(chart)
        props = self._ensure_dashboard_props()
        props.chartEdges = [
            edge
            for edge in (props.chartEdges or [])
            if edge.destinationUrn != str(chart_urn)
        ]
        self._set_aspect(props)

    @property
    def dashboards(self) -> List[DashboardUrn]:
        """Get the dashboards of the dashboard."""
        props = self._ensure_dashboard_props()
        return [
            DashboardUrn.from_string(dashboard.destinationUrn)
            for dashboard in (props.dashboards or [])
        ]

    def set_dashboards(
        self, dashboards: List[Union[DashboardUrnOrStr, Dashboard]]
    ) -> None:
        """Set the dashboards of the dashboard."""
        props = self._ensure_dashboard_props()
        for dashboard in dashboards:
            if isinstance(dashboard, Dashboard):
                dashboard_urn = dashboard.urn
            else:
                dashboard_urn = DashboardUrn.from_string(dashboard)
            props.dashboards.append(models.EdgeClass(destinationUrn=str(dashboard_urn)))
        self._set_aspect(props)

    def add_dashboard(self, dashboard: Union[DashboardUrnOrStr, Dashboard]) -> None:
        """Add a dashboard to the dashboard."""
        if isinstance(dashboard, Dashboard):
            dashboard_urn = dashboard.urn
        else:
            dashboard_urn = DashboardUrn.from_string(dashboard)
        props = self._ensure_dashboard_props()
        dashboards = props.dashboards or []
        existing_urns = [dashboard.destinationUrn for dashboard in dashboards]
        if str(dashboard_urn) not in existing_urns:
            dashboards.append(models.EdgeClass(destinationUrn=str(dashboard_urn)))



import logging
from datetime import datetime
from typing import Dict, Iterable, List, Optional, Tuple, cast

from dateutil.relativedelta import relativedelta

from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.bigquery_v2.bigquery_audit import BigqueryTableIdentifier
from datahub.ingestion.source.bigquery_v2.bigquery_config import BigQueryV2Config
from datahub.ingestion.source.bigquery_v2.bigquery_report import BigQueryV2Report
from datahub.ingestion.source.bigquery_v2.bigquery_schema import (
    RANGE_PARTITION_NAME,
    BigqueryTable,
)
from datahub.ingestion.source.sql.sql_generic import BaseTable
from datahub.ingestion.source.sql.sql_generic_profiler import (
    GenericProfiler,
    TableProfilerRequest,
)
from datahub.ingestion.source.state.profiling_state_handler import ProfilingHandler

logger = logging.getLogger(__name__)


class BigqueryProfiler(GenericProfiler):
    config: BigQueryV2Config
    report: BigQueryV2Report

    def __init__(
        self,
        config: BigQueryV2Config,
        report: BigQueryV2Report,
        state_handler: Optional[ProfilingHandler] = None,
    ) -> None:
        super().__init__(config, report, "bigquery", state_handler)
        self.config = config
        self.report = report

    @staticmethod
    def get_partition_range_from_partition_id(
        partition_id: str, partition_datetime: Optional[datetime]
    ) -> Tuple[datetime, datetime]:
        partition_range_map: Dict[int, Tuple[relativedelta, str]] = {
            4: (relativedelta(years=1), "%Y"),
            6: (relativedelta(months=1), "%Y%m"),
            8: (relativedelta(days=1), "%Y%m%d"),
            10: (relativedelta(hours=1), "%Y%m%d%H"),
        }

        duration: relativedelta
        if partition_range_map.get(len(partition_id)):
            (delta, format) = partition_range_map[len(partition_id)]
            duration = delta
            if not partition_datetime:
                partition_datetime = datetime.strptime(partition_id, format)
            else:
                partition_datetime = datetime.strptime(
                    partition_datetime.strftime(format), format
                )

        else:
            raise ValueError(
                f"check your partition_id {partition_id}. It must be yearly/monthly/daily/hourly."
            )
        upper_bound_partition_datetime = partition_datetime + duration
        return partition_datetime, upper_bound_partition_datetime

    def generate_partition_profiler_query(
        self,
        project: str,
        schema: str,
        table: BigqueryTable,
        partition_datetime: Optional[datetime] = None,
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Method returns partition id if table is partitioned or sharded and generate custom partition query for
        partitioned table.
        See more about partitioned tables at https://cloud.google.com/bigquery/docs/partitioned-tables
        """
        logger.debug(
            f"generate partition profiler query for project: {project} schema: {schema} and table {table.name}, partition_datetime: {partition_datetime}"
        )
        partition = table.max_partition_id
        if table.partition_info and partition:
            partition_where_clause: str

            if table.partition_info.type == RANGE_PARTITION_NAME:
                if table.partition_info.column:
                    partition_where_clause = (
                        f"{table.partition_info.column.name} >= {partition}"
                    )
                else:
                    logger.warning(
                        f"Partitioned table {table.name} without partition column"
                    )
                    self.report.profiling_skipped_invalid_partition_ids[
                        f"{project}.{schema}.{table.name}"
                    ] = partition
                    return None, None
            else:
                logger.debug(
                    f"{table.name} is partitioned and partition column is {partition}"
                )
                try:
                    (
                        partition_datetime,
                        upper_bound_partition_datetime,
                    ) = self.get_partition_range_from_partition_id(
                        partition, partition_datetime
                    )
                except ValueError as e:
                    logger.error(
                        f"Unable to get partition range for partition id: {partition} it failed with exception {e}"
                    )
                    self.report.profiling_skipped_invalid_partition_ids[
                        f"{project}.{schema}.{table.name}"
                    ] = partition
                    return None, None

                partition_data_type: str = "TIMESTAMP"
                # Ingestion time partitioned tables has a pseudo column called _PARTITIONTIME
                # See more about this at
                # https://cloud.google.com/bigquery/docs/partitioned-tables#ingestion_time
                partition_column_name = "_PARTITIONTIME"
                if table.partition_info.column:
                    partition_column_name = table.partition_info.column.name
                    partition_data_type = table.partition_info.column.data_type
                if table.partition_info.type in ("HOUR", "DAY", "MONTH", "YEAR"):
                    partition_where_clause = f"`{partition_column_name}` BETWEEN {partition_data_type}('{partition_datetime}') AND {partition_data_type}('{upper_bound_partition_datetime}')"
                else:
                    logger.warning(
                        f"Not supported partition type {table.partition_info.type}"
                    )
                    self.report.profiling_skipped_invalid_partition_type[
                        f"{project}.{schema}.{table.name}"
                    ] = table.partition_info.type
                    return None, None
            custom_sql = """
SELECT
    *
FROM
    `{table_catalog}.{table_schema}.{table_name}`
WHERE
    {partition_where_clause}
            """.format(
                table_catalog=project,
                table_schema=schema,
                table_name=table.name,
                partition_where_clause=partition_where_clause,
            )

            return (partition, custom_sql)
        elif table.max_shard_id:
            # For sharded table we want to get the partition id but not needed to generate custom query
            return table.max_shard_id, None

        return None, None

    def get_workunits(
        self, project_id: str, tables: Dict[str, List[BigqueryTable]]
    ) -> Iterable[MetadataWorkUnit]:
        profile_requests: List[TableProfilerRequest] = []

        for dataset in tables:
            for table in tables[dataset]:
                normalized_table_name = BigqueryTableIdentifier(
                    project_id=project_id, dataset=dataset, table=table.name
                ).get_table_name()

                if table.external and not self.config.profiling.profile_external_tables:
                    self.report.profiling_skipped_other[f"{project_id}.{dataset}"] += 1
                    logger.info(
                        f"Skipping profiling of external table {project_id}.{dataset}.{table.name}"
                    )
                    continue

                # Emit the profile work unit
                logger.debug(
                    f"Creating profile request for table {normalized_table_name}"
                )
                profile_request = self.get_profile_request(table, dataset, project_id)
                if profile_request is not None:
                    self.report.report_entity_profiled(profile_request.pretty_name)
                    profile_requests.append(profile_request)
                else:
                    logger.debug(
                        f"Table {normalized_table_name} was not eliagible for profiling."
                    )

        if len(profile_requests) == 0:
            return
        yield from self.generate_profile_workunits(
            profile_requests,
            max_workers=self.config.profiling.max_workers,
            platform=self.platform,
            profiler_args=self.get_profile_args(),
        )

    def get_dataset_name(self, table_name: str, schema_name: str, db_name: str) -> str:
        return BigqueryTableIdentifier(
            project_id=db_name, dataset=schema_name, table=table_name
        ).get_table_name()

    def get_batch_kwargs(
        self, table: BaseTable, schema_name: str, db_name: str
    ) -> dict:
        return dict(
            schema=db_name,  # <project>
            table=f"{schema_name}.{table.name}",  # <dataset>.<table>
        )

    def get_profile_request(
        self, table: BaseTable, schema_name: str, db_name: str
    ) -> Optional[TableProfilerRequest]:
        profile_request = super().get_profile_request(table, schema_name, db_name)

        if not profile_request:
            return None

        # Below code handles profiling changes required for partitioned or sharded tables
        # 1. Skip profile if partition profiling is disabled.
        # 2. Else update `profile_request.batch_kwargs` with partition and custom_sql

        bq_table = cast(BigqueryTable, table)
        (partition, custom_sql) = self.generate_partition_profiler_query(
            db_name, schema_name, bq_table, self.config.profiling.partition_datetime
        )

        if partition is None and bq_table.partition_info:
            self.report.report_warning(
                title="Profile skipped for partitioned table",
                message="profile skipped as partitioned table is empty or partition id or type was invalid",
                context=profile_request.pretty_name,
            )
            return None
        if (
            partition is not None
            and not self.config.profiling.partition_profiling_enabled
        ):
            logger.debug(
                f"{profile_request.pretty_name} and partition {partition} is skipped because profiling.partition_profiling_enabled property is disabled"
            )
            self.report.profiling_skipped_partition_profiling_disabled.append(
                profile_request.pretty_name
            )
            return None

        if partition:
            logger.debug("Updating profiling request for partitioned/sharded tables")
            profile_request.batch_kwargs.update(
                dict(
                    custom_sql=custom_sql,
                    partition=partition,
                )
            )

        return profile_request
